/*******************************************

 MAIN.CPP - The KINT Interpreter
	Copyright 2002-2003 Everett Kaser
	All rights reserved.

	Do not copy or redistribute without the
	author's explicit written permission.
 *******************************************

The KINT 2.0 interpreter is divided into two .C files and two matching .H files.
The two .C files are pretty much divided into mostly OS-dependant code (MAIN.C)
and OS-indendent code (PMACH.C).  Of course, there's going to be some cross-over
between, particularly depending upon the "native bitmap data format" on the
target machine.  With 2.0, KINT does ALL of it's own drawing of graphics, including
bitmaps, points, lines, rectangles, ovals, and text.  This is all done to memory
arrays that represent the bitmap data, and there is one of these that represents
the game's display output at any given time (kBuf).  The bitmap data is arranged
(and thus manipulated by the routines in PMACH.C) such that it's consistent with
Microsoft Window's DIB bitmaps, and thus can be easily pointed to and BitBlt'd
to the actual screen during the WM_PAINT message (any time the OS asks for the
game window to be redrawn).  If you want to port KINT to another OS, its native
bitmap format may be different, in which case it would be more efficient (in
terms of code execution) if those routines in PMACH.C were modified so that the
KINT memory bitmap data format matched your target OS's, and the routines that
drew into it would, of course, then have to be modified to read/write that different
format.

The KINT 2.0 interpreter has added some significant functionality to the graphics,
including clipping and rotation of bitmaps.  The bitmaps don't actually get
rotated, but rather all of the drawing routines access the bitmap memory blocks
(BMBs) via pointers and global variables that get set up at the beginning of
each routine based upon the current rotation state of the source and target
bitmaps (see SetupBMD() and SetupBMS() in PMACH.C).  These routines set a pointer
to the pixel that in the BMB that is the LOGICAL top-left corner (if the bitmap
werer NOT rotated).  So, for example, if the bitmap is rotated 180 degrees, then
this pointer would be set to the PHYSICAL lower-right pixel of the bitmap.  Further,
there are variables that give the amount (offset) to add to the pointer in order
to step one pixel in the X direction and one pixel in the Y direction.  Similarly,
a translation process is done in MAIN.C to convert mouse-pointer coordinates from
the physical screen coordinates to properly "rotated" coordinates (see FixXY() in
MAIN.C) and to convert keyboard cursor keys (see MapCursor() in MAIN.C) so that
everything works relatively transparently for the KASM code itself (ie, the KASM
.KE program doesn't have to know or care which way the bitmaps are rotated, although
it CAN find out using the SysSizeBM opcode and can rotate bitmaps using the SysOrientBM
opcode).  The BMB data structure has physW/physH and logW/logH values.  The phys values
are the actual width and height of the bitmap as it was allocated in RAM, whereas the
log values are the rotated-values which are dependant upon the bitmap's current
orientation at any given moment.

There is also the ability to have the graphics "double-sized".  This was
done it make developing Pocket PC applications on a desktop a little easier, since
a 320x240 window on a 1280x1024 display is pretty dinky.  Normally on a desktop
interpreter, you'd want KINT_DOUBLE_SIZE to be 0.  Of course, that double-sizing
effects the display OUTPUT as well as the mouse INPUT.  There is also a KINT_POCKET_PC_EMUL
#define that controls whether the interpreter is trying to emulate a Pocket PC
environment on a desktop screen or not.  ie, these are the supported combinations:
                   KINT_DESKTOP_WINDOWS KINT_POCKET_PC  KINT_POCKET_PC_EMUL  KINT_DOUBLE_SIZE
desktop						1				  0					  0					0
Pocket PC					0				  1					  0					0
desktop Pocket PC			1				  0					  1					0
desktop Pocket PC			1				  0					  1					1

See MAIN.H for all of these #define's.

There is a lot of #if'd code sprinkled throughout MAIN.C to be able to generate a
KINT 2.0 interpreter for either the desktop MS Windows OS's OR for the Pocket PC
devices running Microsoft Windows CE OS.  If you're porting this to a completely
different OS, the simplest thing might be to go through the code and hack out the
#if'd code that doesn't match your OS very well (ie, if you're porting to another
handheld, you'd get rid of any KINT_DESKTOP_WINDOWS code, and if you're porting to
another desktop OS, you'd get rid of any KINT_POCKET_PC code).

There is some UNICODE stuff.  The KASM code only deals with 8-bit characters.
However, the Pocket PC uses ONLY Unicode for file names, and other OS's may
move to that in the future.  Hence, the Widen() and Unwiden() functions to 
convert back and forth between 8 and 16 bit characters.

The KHDR_IDLE message that is being passed in the main message-loop: it's not
currently (and probably will remain) not terribly important.  I can't guarantee
that, but I try to put all "periodic activity" on the KHDR_TIMER call.  So, if
your OS structure makes the KHDR_IDLE call difficult to implement, you can PROBABLY
get by without it (but I make no guarantees).

The code is not as clean and neat as I'd like, but time, reality, and the desire for
fun rather than drudgery all exert a great deal of influence on the final appearance.

 ************************************************************/

//
// Mac nice-to-have:
//
// - Hero's Friends writes bmp data to the game folder.
//   Fix so that it writes elsewhere and the game can still find it.
//
// - Find way to disable/hide the shareware "make a copy" buttons/UI.
//

#define DEBUG_FILE 0
#define DEBUG_KINT_EVENTS 0
#define DEBUG_MAC_EVENTS 1
#define DEBUG_SOUND 0
#define DEBUG_PRINT 0
#define DEBUG_UPDATES 1

#define USE_HIVIEW 1
#define USE_CMD_KEY 1

// The idle timer period, in milliseconds
#define IDLE_MS 100

#define FAKE_SCREEN_SIZE	0
#define FAKE_SCREEN_SIZE_W	800
#define FAKE_SCREEN_SIZE_H	600

/*********************  Header Files  *********************/

#include <CoreFoundation/CoreFoundation.h>
#include <QuickTime/QuickTime.h>

#ifdef __cplusplus
extern "C" {
#endif
#include <mach/mach_time.h>
#include "main.h"
#ifdef __cplusplus
}
#endif

#include <cstdio>
#include <cstdlib>
#include <string>
#include <vector>
#include <deque>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>

#if BETA
  #include <time.h>
#endif

using std::string;

/*********************  Local Structures  *********************/

typedef struct
{
	Movie		mMovie;
	Handle		mDataRef;
	bool		mLooping;
	bool		mPlaying;
	bool		mAutoDelete;
	bool		mIsMusic;
	QTCallBack	qtCallback;
} movieSound_t;
static std::deque<movieSound_t *> sSoundOrphans;

class KINTmsg
{
public:
	WORD mCmd;
	DWORD mArg1, mArg2, mArg3;

	KINTmsg (WORD cmd, DWORD arg1, DWORD arg2, DWORD arg3)
	{
		mCmd = cmd; mArg1 = arg1; mArg2 = arg2; mArg3 = arg3;
	}
	
	~KINTmsg () {}
};
std::deque<KINTmsg *> KINTMsgQueue;

typedef struct
{
	int width;
	int height;
} windowSizes_t;

//**** FindFirst/FindNext classes
class FileEntry
{
public:
	BYTE	attributes;
	std::string	name;
	BOOL	AppData;

	FileEntry(const char *inName, BYTE inAttributes, BOOL inAppData);
	~FileEntry(void) {}
};

typedef std::vector<FileEntry *>::iterator fileListIter_t;

class FindFileList
{
public:
	std::vector<FileEntry *> mList;
	fileListIter_t iter;

	FindFileList(void) {}
	~FindFileList(void);
};


FindFileList::~FindFileList(void)
{
	fileListIter_t iter;
	
	for(iter = mList.begin(); iter != mList.end(); iter++) delete *iter;
};


FileEntry::FileEntry(const char *inName, BYTE inAttributes, BOOL inAppData)
{
	name = inName;
	attributes = inAttributes;
	AppData = inAppData;
}


/*********************  Local Prototypes  *********************/

int wildicmp(const char *wild, const char *string);
bool CreateMacWindow (void);
void BuildPath(const char *pMem, BYTE inFolderType, BOOL AppData);
void PlaySoundFromFile (const char *inFilePath);
void StopSound_QT (movieSound_t *inSound);
static int ConvertCFStringToUnicode (const CFStringRef inString, UniChar *outString);
void Kff(BOOL appdata, BOOL removeExtension);

//********************************************************************
extern PBMB		kBuf, pbFirst;

//************************************************************
// pmach GLOBALS
int		PhysicalScreenW, PhysicalScreenH, PhysicalScreenSizeIndex;
WORD	MusicPlaying;
//int		CurCur;
KE		mainKE;

/*******************  Global Variables ********************/

CFStringRef gAppName;

WindowRef gMacWindow;
CGContextRef gCGContext;

bool gInBackground;
bool sIgnoreNextMouseUp;
volatile bool gInPrintLoop;
bool gUpdateImage;
UInt32 gOSXVersion32;

EventLoopTimerRef sTimerRef = NULL;
EventLoopTimerRef sIdleTimerRef = NULL;
DWORD sTimerPeriod;
BOOL inSM=FALSE;

BOOL InSetWindowPos;
BOOL ProgramReady;
BOOL WinHide = false;
BOOL AppHide = false;
BOOL DockHide = false;

char	OSctrl_cmdSTR[32] = "Cmd";
TCHAR	PgmPath[PATHLEN], TmpPath[PATHLEN], AppDataPath[PATHLEN], AppPath[PATHLEN];
BOOL	CopyAppData=TRUE;
struct	stat	fstats;
FindFileList	*FirstNextList=NULL;
UInt32	lastModifiers;
WORD	MouseButtonsDown=0;

windowSizes_t windowSizes[] = {
	{ 640, 480 },
	{ 800, 600 },
	{ 1024, 768 },
	{ 1152, 864 },
	{ 1280, 1024 },
	{ 1600, 1200 },
	{ 60000, 60000 }
};


#pragma mark -

// KINT requires VK key codes. This table will take a Mac keyCode and remap it to the closest VK equivalent.
static BYTE sMacKeyCodeToVK[128] =
{
	0x41, 0x53, 0x44, 0x46, 0x48, 0x47, 0x5a, 0x58, 0x43, 0x56, 0x00, 0x42, 0x51, 0x57, 0x45, 0x52,
	0x59, 0x54, 0x31, 0x32, 0x33, 0x34, 0x36, 0x35, 0xbb, 0x39, 0x37, 0xbd, 0x38, 0x30, 0xdd, 0x4f,
	0x55, 0xdb, 0x49, 0x50, 0x0d, 0x4c, 0x4a, 0xde, 0x4b, 0xba, 0xdc, 0xbc, 0xbf, 0x4e, 0x4d, 0xbe,
	0x09, 0x20, 0xc0, 0x08, 0x00, 0x1b, 0x00, 0x00, 0x10, 0x14, 0x12, 0x11, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x6e, 0x00, 0x6a, 0x00, 0x6b, 0x00, 0x90, 0x00, 0x00, 0x00, 0x6f, 0x0d, 0x00, 0x6d, 0x00,
	0x00, 0xbb, 0x60, 0x61, 0x62, 0x63, 0x64, 0x65, 0x66, 0x67, 0x00, 0x68, 0x69, 0x00, 0x00, 0x00,
	0x74, 0x75, 0x76, 0x72, 0x77, 0x78, 0x00, 0x7a, 0x00, 0x2c, 0x00, 0x91, 0x00, 0x79, 0x00, 0x7b,
	0x00, 0x13, 0x2d, 0x24, 0x21, 0x2e, 0x73, 0x23, 0x71, 0x22, 0x70, 0x25, 0x27, 0x28, 0x26, 0x00
};

// These are used by the modifier-changed events
enum
{
	VK_SHIFT	= 0x10,
	VK_CONTROL	= 0x11,
	VK_MENU		= 0x12,
	VK_CAPITAL	= 0x14
};

BOOL OtherError=FALSE;

//********************************************************************
void AddMsg(WORD cmd, DWORD arg1, DWORD arg2, DWORD arg3)
{
	if (!gMacWindow || !sIdleTimerRef)
	{
#if _DEBUG && DEBUG_KINT_EVENTS
		printf("AddMsg cmd: %d %d %d %d\n", cmd, arg1, arg2, arg3);
#endif
		KEmsg(&mainKE, cmd, arg1, arg2, arg3);
	}
	else
	{
		KINTmsg *newMsg = new KINTmsg(cmd, arg1, arg2, arg3);
		KINTMsgQueue.push_back(newMsg);

		// Wake up the idle timer so it can process the KINT events immediately
		SetEventLoopTimerNextFireTime(sIdleTimerRef, kEventDurationNoWait);
	}
}

//********************************************************************
void SendNextMsg()
{
	if (!KINTMsgQueue.empty())
	{
		KINTmsg *kintMsg = KINTMsgQueue.front();
#if _DEBUG && DEBUG_KINT_EVENTS
		printf("SendNextMsg cmd: %d %d %d %d\n", kintMsg->mCmd, kintMsg->mArg1, kintMsg->mArg2, kintMsg->mArg3);
#endif
		KEmsg (&mainKE, kintMsg->mCmd, kintMsg->mArg1, kintMsg->mArg2, kintMsg->mArg3);
		KINTMsgQueue.pop_front();
		delete kintMsg;
	}
}

void MacNotImplemented(void)
{
#if _DEBUG
	Debugger();
#endif
}

bool MacDoQuit(void)
{
	if( KEmsg(&mainKE, KHDR_QUERY_QUIT, 0, 0, 0) )
	{
		KEmsg(&mainKE, KHDR_SHUT_DOWN, 0, 0, 0);
		return true;
	}
	return false;
}

pascal OSStatus appCommandHandler (EventHandlerCallRef handlerRef, EventRef inEvent, void *userData)
{
    OSStatus status = eventNotHandledErr;
	UInt32 eventClass = GetEventClass (inEvent);
	UInt32 eventKind = GetEventKind (inEvent);
	
	if (eventClass == kEventClassCommand && eventKind == kEventProcessCommand)
	{
		HICommand command; 
		GetEventParameter (inEvent, kEventParamDirectObject, typeHICommand, NULL, sizeof(HICommand), NULL,  &command);

		switch (command.commandID)
		{
			case kHICommandPreferences:
				break;

			case kHICommandQuit:
				MacDoQuit();
				status = noErr;
				break;

			default:
				status = eventNotHandledErr;
		}
	}
	else if (eventClass == kEventClassApplication)
	{
		switch (eventKind)
		{
			case kEventAppActivated:
				gInBackground = false;
				break;
			case kEventAppDeactivated:
				gInBackground = true;
				break;
			case kEventAppHidden:
				AddMsg(KHDR_WINDOW_HIDDEN, 1, 0, 0);
				AppHide = true;
				break;
			case kEventAppShown:
				if (AppHide)
					AddMsg(KHDR_WINDOW_HIDDEN, 0, 0, 0);
				AppHide = false;
				break;
		}
	}
    return status;
}

pascal OSStatus windowEventHandler (EventHandlerCallRef inCaller, EventRef inEvent, void* inUserData)
{
	OSStatus result = eventNotHandledErr;

	UInt32 eventClass = GetEventClass (inEvent);
	UInt32 eventKind = GetEventKind (inEvent);
	
	SetPortWindowPort (gMacWindow);

	if (eventClass == kEventClassWindow)
	{
		WindowRef macWindow;
		(void)GetEventParameter(inEvent, kEventParamDirectObject, typeWindowRef, NULL, sizeof(macWindow), NULL, &macWindow);

		switch (eventKind)
		{
#if !USE_HIVIEW
			case kEventWindowUpdate:
				{
					BeginUpdate(macWindow);
					CGImageRef image = CGBitmapContextCreateImage (kBuf->bitmapContext);
					
					// We want to draw the bitmap in the upper-left of the content region of the window.
					
					// CGContextDrawImage will scale the CGImage by cgRect. As it happens, our window will
					// be smaller than the bitmap rect, so we must use the bitmap rect instead of the window
					// rect to avoid any scaling of the image.
					
					// We also need to compensate for CG's inverted coordinate system, so we subtract
					// the image height from the window content region height to offset the y rendering
					// of the CGBitmap.
					
					Rect bounds;
					GetWindowPortBounds(gMacWindow, &bounds);
					int yOffset = (bounds.bottom - bounds.top) - CGImageGetHeight(image);
					CGRect cgRect = CGRectMake(0, yOffset, CGImageGetWidth(image),  CGImageGetHeight(image));
					
					CGContextDrawImage (gCGContext, cgRect, image);
					CGContextFlush(gCGContext);
					CGImageRelease (image);
					EndUpdate(macWindow);
				}
				break;
#endif
				
			// The window is going to/from the dock (no size change)
			case kEventWindowCollapsed:
				AddMsg(KHDR_WINDOW_HIDDEN, 1, 0, 0);
				DockHide = true;
				break;
			case kEventWindowExpanded:
				if( DockHide ) AddMsg(KHDR_WINDOW_HIDDEN, 0, 0, 0);
				DockHide = false;
				break;
			case kEventWindowHidden:
				AddMsg(KHDR_WINDOW_HIDDEN, 1, 0, 0);
				WinHide = true;
				break;
			case kEventWindowShown:
				// KINT doesn't like receiving unhide messages at startup when the window is first shown.
				if( WinHide == true ) AddMsg(KHDR_WINDOW_HIDDEN, 0, 0, 0);

				if( !InSetWindowPos ) {
					if( kBuf->logW > kBuf->physW || kBuf->logH > kBuf->physH ) InitScreen(kBuf->logW, kBuf->logH);
					AddMsg(KHDR_WINDOW_SIZE, kBuf->logW, kBuf->logH, K_MAXIMIZED);
				}
				WinHide = false;
				break;
			case kEventWindowBoundsChanged:
				{
					UInt32 attribs;

					(void)GetEventParameter(inEvent, kEventParamAttributes, typeUInt32, NULL, sizeof(attribs), NULL, &attribs);

					if (attribs & kWindowBoundsChangeSizeChanged) {

						Rect newBounds;
						
						(void)GetEventParameter(inEvent, kEventParamCurrentBounds, typeQDRectangle, NULL, sizeof(newBounds), NULL, &newBounds);
						if( kBuf->orient & BMO_90 ) {
							kBuf->logW = newBounds.bottom - newBounds.top;
							kBuf->logH = newBounds.right - newBounds.left;
						} else {
							kBuf->logW = newBounds.right - newBounds.left;
							kBuf->logH = newBounds.bottom - newBounds.top;
						}
						
						if( !InSetWindowPos ) {
							if( kBuf->logW > kBuf->physW || kBuf->logH > kBuf->physH ) InitScreen(kBuf->logW, kBuf->logH);
							AddMsg(KHDR_WINDOW_SIZE, kBuf->logW, kBuf->logH, K_SIZED);
							if( ProgramReady ) SendNextMsg();
						}
					}
					result = noErr;
					return eventNotHandledErr;
				}
				break;

			case kEventWindowClose:
				MacDoQuit();
				result = noErr;
				break;
		}
	}
	else if (eventClass == kEventClassKeyboard)
	{
		char keyChar;
		(void)GetEventParameter(inEvent, kEventParamKeyMacCharCodes, typeChar, NULL, sizeof(keyChar), NULL, &keyChar);
		UInt32 modifiers;
		(void)GetEventParameter(inEvent, kEventParamKeyModifiers, typeUInt32, NULL, sizeof(modifiers), NULL, &modifiers);

		UInt32 keyCode;
		(void)GetEventParameter(inEvent, kEventParamKeyCode, typeUInt32, NULL, sizeof(keyCode), NULL, &keyCode);
		BYTE vkCode = sMacKeyCodeToVK[keyCode & 0x7f];
//char geekBuf[256];
//sprintf(geekBuf, "keyChar=%d modifiers=%d keyCode=%d vkCode=%d", keyChar, modifiers, keyCode, vkCode);
//SystemMessage(geekBuf);		

		switch (eventKind)
		{
			case kEventRawKeyDown:
//SystemMessage("kEventRawKeyDown");
				if( vkCode==0x20 || (vkCode >= 0x30 && vkCode <= 0x39) || (vkCode >= 0x41 && vkCode <= 0x5A) || (vkCode >= 0xBA && vkCode<=0xE0) ) {
//SystemMessage("KHDR_CHAR");
					AddMsg(KHDR_ALPHAKEY_DOWN, vkCode, 1 /* repeatRate */, 0);
					AddMsg(KHDR_CHAR, keyChar, 1 /* repeatRate */, 0);
				} else {
//SystemMessage("KHDR_KEY_DOWN");
					AddMsg(KHDR_KEY_DOWN, vkCode, 1 /* repeatRate */, 0);
				}
				result = noErr;
				break;
				
			case kEventRawKeyUp:
				if( (vkCode >= 0x30 && vkCode <= 0x39) || (vkCode >= 0x41 && vkCode <= 0x5A) || (vkCode >= 0xBA && vkCode<=0xE0) )
					AddMsg(KHDR_ALPHAKEY_UP, vkCode, 1 /* repeatRate */, 0);
				else
					AddMsg(KHDR_KEY_UP, vkCode, 1 /* repeatRate */, 0);
				result = noErr;
				break;

			case kEventRawKeyModifiersChanged:
			{
				WORD msgType = 0;

				UInt32 changed = modifiers ^ lastModifiers;
				lastModifiers = modifiers;
				if (changed & shiftKey) {
					keyChar = VK_SHIFT;
					if( modifiers & shiftKey ) {
						msgType = KHDR_KEY_DOWN;
						MouseButtonsDown |= 0x04;
					} else {
						msgType = KHDR_KEY_UP;
						MouseButtonsDown &= ~0x04;
					}
					AddMsg(msgType, keyChar, 0 /* repeatRate */, 0);
				}
				if (changed & optionKey) {
					keyChar = VK_MENU;
					if( modifiers & optionKey ) {
						msgType = KHDR_KEY_DOWN;
						MouseButtonsDown |= 0x8000;
					} else {
						msgType = KHDR_KEY_UP;
						MouseButtonsDown &= ~0x8000;
					}
					AddMsg(msgType, keyChar, 0 /* repeatRate */, 0);
				}
#if USE_CMD_KEY
				if (changed & cmdKey) {
					keyChar = VK_CONTROL;
					if( modifiers & cmdKey ) {
						msgType = KHDR_KEY_DOWN;
						MouseButtonsDown |= 0x08;
					} else {
						msgType = KHDR_KEY_UP;
						MouseButtonsDown &= ~0x08;
					}
					AddMsg(msgType, keyChar, 0 /* repeatRate */, 0);
				}
#else
				if (changed & controlKey) {
					keyChar = VK_CONTROL;
					if( modifiers & controlKey ) {
						msgType = KHDR_KEY_DOWN;
						MouseButtonsDown |= 0x08;
					} else {
						msgType = KHDR_KEY_UP;
						MouseButtonsDown &= ~0x08;
					}
					AddMsg(msgType, keyChar, 0 /* repeatRate */, 0);
				}
#endif
				result = noErr;
			}
		}
	}
	else if (eventClass == kEventClassMouse)
	{
		WindowRef macWindow;
		(void)GetEventParameter(inEvent, kEventParamWindowRef, typeWindowRef, NULL, sizeof(macWindow), NULL, &macWindow);
		Point where;
		(void)GetEventParameter (inEvent, kEventParamWindowMouseLocation, typeQDPoint, NULL, sizeof(where), NULL, &where);
		WindowPartCode	partCode;
		(void)GetEventParameter(inEvent, kEventParamWindowPartCode, typeWindowPartCode, NULL, sizeof(partCode), NULL, &partCode);
		Rect structRect;
		(void)GetWindowStructureWidths (macWindow, &structRect);
		where.h -= structRect.left;
		where.v -= structRect.top;
		DWORD location = (where.h & 0xffff) | ((where.v & 0xffff) << 16);

		EventMouseButton theButton;
		(void)GetEventParameter (inEvent, kEventParamMouseButton, typeMouseButton, NULL, sizeof(theButton), NULL, &theButton);
		
		switch (eventKind)
		{
			case kEventMouseDown:
				if (gInBackground || gInPrintLoop)
					sIgnoreNextMouseUp = true;
				else if (where.v >= 0)
				{
					switch (theButton)
					{
						case 1:
							if( partCode==inGrow ) return eventNotHandledErr;
#if USE_CMD_KEY
	// If we're remapping the Ctrl key to the Cmd key, force a Ctrl-left-click to emulate a right-click
							if( lastModifiers & controlKey ) goto down2;
#endif
							MouseButtonsDown |= 0x01;
							AddMsg(KHDR_KEY_DOWN, 0x01, MouseButtonsDown, location);
							break;
						case 2:
down2:
							MouseButtonsDown |= 0x02;
							AddMsg(KHDR_KEY_DOWN, 0x02, MouseButtonsDown, location);
							break;
						case 3:
							// Map button 3 to what KINT expects (VK_MBUTTON)
							MouseButtonsDown |= 0x10;
							AddMsg(KHDR_KEY_DOWN, 0x04, MouseButtonsDown, location);
							break;
						case 4:
							// Map button 4 to KINT's "Back"
							MouseButtonsDown |= 0x20;
							AddMsg(KHDR_KEY_DOWN, 0x00008001, MouseButtonsDown, location);
							break;
						case 5:
							// Map button 5 to KINT's "Next"
							MouseButtonsDown |= 0x40;
							AddMsg(KHDR_KEY_DOWN, 0x00008002, MouseButtonsDown, location);
							break;
						default:
							// Ignore any buttons > 5. KINT doesn't use them.
							break;
					}
					result = noErr;
				}
				break;
			case kEventMouseUp:
				// If we got a mouse-down while in the background, ignore the matching
				// mouse-up. This avoids sending mouse-up events to KINT when a mouse-click
				// was used to move the app to the foreground.
				if (sIgnoreNextMouseUp)
					sIgnoreNextMouseUp = false;
				else if (where.v >= 0 && !gInBackground)
				{
					switch (theButton)
					{
						case 1:
#if USE_CMD_KEY
	// If we're remapping the Ctrl key to the Cmd key, force a Ctrl-left-click to emulate a right-click
							if( lastModifiers & controlKey ) goto up2;
#endif
							MouseButtonsDown &= ~0x01;
							AddMsg(KHDR_KEY_UP, 0x01, MouseButtonsDown, location);
							break;
						case 2:
up2:
							MouseButtonsDown &= ~0x02;
							AddMsg(KHDR_KEY_UP, 0x02, MouseButtonsDown, location);
							break;
						case 3:
							// Map button 3 to what KINT expects (VK_MBUTTON)
							MouseButtonsDown &= ~0x10;
							AddMsg(KHDR_KEY_UP, 0x04, MouseButtonsDown, location);
							break;
						case 4:
							// Map button 4 to KINT's "Back"
							MouseButtonsDown &= ~0x20;
							AddMsg(KHDR_KEY_UP, 0x00008001, MouseButtonsDown, location);
							break;
						case 5:
							// Map button 5 to KINT's "Next"
							MouseButtonsDown &= ~0x40;
							AddMsg(KHDR_KEY_UP, 0x00008002, MouseButtonsDown, location);
							break;
						default:
							// Ignore any buttons > 5. KINT doesn't use them.
							break;
					}
					result = noErr;
				}
				break;
			case kEventMouseMoved:
				AddMsg(KHDR_POINTER_MOVE, where.h, where.v, MouseButtonsDown);
				result = noErr;
				break;
			case kEventMouseDragged:
				AddMsg(KHDR_POINTER_MOVE, where.h, where.v, MouseButtonsDown);//theButton);
				result = noErr;
				break;
			case kEventMouseWheelMoved:
				{
					EventMouseWheelAxis mouseAxis;
					(void)GetEventParameter (inEvent, kEventParamMouseWheelAxis, typeMouseWheelAxis, NULL, sizeof(mouseAxis), NULL, &mouseAxis);
					
					SInt32 mouseDelta;
					(void)GetEventParameter (inEvent, kEventParamMouseWheelDelta, typeSInt32, NULL, sizeof(mouseDelta), NULL, &mouseDelta);
					
					if (mouseAxis == kEventMouseWheelAxisY)
						AddMsg(KHDR_KEY_DOWN, 0x00008003, mouseDelta, location);
					result = noErr;
				}
				break;
		}
	}

	return result;
}
DEFINE_ONE_SHOT_HANDLER_GETTER( windowEventHandler );

#if USE_HIVIEW
OSStatus drawEventHandler(EventHandlerCallRef myHandler, EventRef event, void* userData)
{
#if _DEBUG && DEBUG_MAC_EVENTS
	printf ("drawEventHandler\n");
#endif
	CGContextRef cgContext;
	OSStatus status = GetEventParameter(event,
					kEventParamCGContextRef,
					typeCGContextRef,
					NULL,
					sizeof(CGContextRef),
					NULL,
					&cgContext);
	HIRect bounds;
	status = HIViewGetBounds((HIViewRef) userData, &bounds);
	
	// switch to quartz coordinates.
	CGContextTranslateCTM (cgContext, 0, bounds.size.height);
	CGContextScaleCTM (cgContext, 1.0, -1.0);

	if (gUpdateImage)
	{
		if (kBuf->cgImage)
			CGImageRelease (kBuf->cgImage);
		kBuf->cgImage = CGBitmapContextCreateImage (kBuf->bitmapContext);
		gUpdateImage = false;
	}
	// We want to draw the bitmap in the upper-left of the content region of the window.
	
	// CGContextDrawImage will scale the CGImage by cgRect. As it happens, our window will
	// be smaller than the bitmap rect, so we must use the bitmap rect instead of the window
	// rect to avoid any scaling of the image.
	
	// We also need to compensate for CG's inverted coordinate system, so we subtract
	// the image height from the window content region height to offset the y rendering
	// of the CGBitmap.
	
	int yOffset = (bounds.size.height) - CGImageGetHeight(kBuf->cgImage);
	CGRect cgRect = CGRectMake(0, yOffset, CGImageGetWidth(kBuf->cgImage),  CGImageGetHeight(kBuf->cgImage));
	
	CGContextDrawImage (cgContext, cgRect, kBuf->cgImage);

	return status;
}
#endif

//********************************************************************

void KINT_ProcessAllEvents(void)
{
	while (!KINTMsgQueue.empty())
		SendNextMsg();
}

static pascal void sIdleTimerProc(EventLoopTimerRef inTimerRef, EventLoopIdleTimerMessage inAction, void *inUserData)
{
	// We want to keep track of sIdleCount across all calls, so it's static for now.
	static int sIdleCount;
	
	// Don't do any idle functions if we're in the print loop; KINT won't deal with them properly.
	if( gInPrintLoop || inSM ) return;

	KINT_ProcessAllEvents();

	// If any sounds have stopped playing, dispose of them.
	while (!sSoundOrphans.empty())
	{
		movieSound_t *sound = sSoundOrphans.front();
		StopSound_QT(sound);
		sSoundOrphans.pop_front();
	}
	
	// if the timer has just fired, reset the idle count to 0
	if (inAction == kEventLoopIdleTimerStarted) sIdleCount = 0;

	// Call KINT to process any and all IDLE needs. It will return 0 when it's done.
	while ( KEmsg(&mainKE, KHDR_IDLE, sIdleCount++, 0, 0) ) {}
}

void StringFindReplace (string& source, const string findText, const string replaceText)
{
	if (findText.size () > 0)
	{
		string::size_type pos = 0;
		string::size_type found;
		
		while ((found = source.find(findText, pos)) != string::npos)
		{
			source.replace (found, findText.size(), replaceText);
			pos = found + replaceText.size();
		}
	}
}

extern "C" BYTE * MacStringFixup (BYTE *inString)
{
#if USE_CMD_KEY
	static string tempString;
	
	tempString = (char *)inString;
	StringFindReplace (tempString, "Ctl", "Cmd");
	StringFindReplace (tempString, "CTL", "CMD");
	StringFindReplace (tempString, "Ctrl", "Cmd");
	StringFindReplace (tempString, "CTRL", "CMD");
	return (BYTE *)tempString.c_str();
#else
	return inString;
#endif
}

void SetAppDataPath() {

	HFSUniStr255 name;
	FSRef appSupportRef;
	FSRef folderRef;
	FSRef eksRef;
	OSStatus err;
	OSErr	oserr;
	CFURLRef parentURL = NULL;
	CFURLRef eksURL = NULL;
	CFURLRef resourceURL = NULL;
	CFURLRef pgmURL = NULL;
	BOOL	success = FALSE;
//char terr[1024];
// setup AppPath[]		
	if( NULL!=(pgmURL = CFBundleCopyBundleURL (CFBundleGetMainBundle())) ) {
		CFURLGetFileSystemRepresentation(pgmURL, true, (UInt8 *)AppPath, sizeof(AppPath));
		CFRelease(pgmURL);
		strcat(AppDataPath, "/");
	}
// setup PgmPath[]		
	if( NULL!=(pgmURL = CFBundleCopyResourcesDirectoryURL(CFBundleGetMainBundle())) ) {
		CFURLGetFileSystemRepresentation(pgmURL, true, (UInt8 *)PgmPath, sizeof(PgmPath));
		CFRelease(pgmURL);
		strcat(PgmPath, "/");
	}

// setup AppDataPath[]
	// Look in ~/Library/Application Support/Everett Kaser Software/<EKSGameName>"
	oserr = FSFindFolder(kUserDomain, kApplicationSupportFolderType, kCreateFolder, &appSupportRef);
//sprintf(terr, "AppPath=%s", AppPath);
//SystemMessage(terr);
//sprintf(terr, "PgmPath=%s", PgmPath);
//SystemMessage(terr);

	if( oserr==noErr ) {	
		// Get the CFURL and FSRef for the "Everett Kaser Software" folder, creating it if necessary.
		name.length = ConvertCFStringToUnicode (CFSTR("Everett Kaser Software"), name.unicode);
		err = FSCreateDirectoryUnicode (&appSupportRef, name.length, name.unicode, kFSCatInfoNone, NULL, &folderRef, NULL, NULL);
		if( err == dupFNErr ) {
			if( NULL==(parentURL = CFURLCreateFromFSRef(NULL, &appSupportRef)) ) goto fail;
			eksURL = CFURLCreateCopyAppendingPathComponent (NULL, parentURL, CFSTR("Everett Kaser Software"), true);
		} else {
			eksURL = CFURLCreateFromFSRef (NULL, &folderRef);
		}
		if( NULL==eksURL ) goto fail;
		
		(void)CFURLGetFSRef(eksURL, &eksRef);

		// Now get the CFURL and FSRef for the app-specific game name portion of the path, creating if necessary.
		name.length = ConvertCFStringToUnicode (gAppName, name.unicode);
		err = FSCreateDirectoryUnicode (&eksRef, name.length, name.unicode, kFSCatInfoNone, NULL, &folderRef, NULL, NULL);

		if( err == dupFNErr ) resourceURL = CFURLCreateCopyAppendingPathComponent (NULL, eksURL, gAppName, true);
		else resourceURL = CFURLCreateFromFSRef (NULL, &folderRef);

		CFURLGetFileSystemRepresentation(resourceURL, true, (UInt8 *)AppDataPath, sizeof(AppDataPath));
		strcat(AppDataPath, "/");
		success = TRUE;
	}
fail:
	if( !success ) {
		strcpy(AppDataPath, PgmPath);
		CopyAppData = FALSE;
	}
//sprintf(terr, "CopyAppData=%d AppDataPath=%s", CopyAppData, AppDataPath);
//SystemMessage(terr);
	
	if( resourceURL ) CFRelease(resourceURL);
	if( eksURL ) CFRelease(eksURL);
	if( parentURL ) CFRelease(parentURL);
}


#pragma mark -

int main(void)
{
#if BETA
	BYTE	bfb[512], bfa;
	DWORD	bfr;
	struct	tm	*bftm;
	time_t	bftime;
#endif


// debug code to force a crash
#if 0
  if( *((BYTE*)0)==0 ) return 0 ;
#endif
	EventTypeSpec windowEvents[] =
	{
#if !USE_HIVIEW
		{ kEventClassWindow, kEventWindowUpdate },
#endif
		{ kEventClassWindow, kEventWindowClose },
		{ kEventClassWindow, kEventWindowCollapsed },
		{ kEventClassWindow, kEventWindowExpanded },
		{ kEventClassWindow, kEventWindowBoundsChanged },
		{ kEventClassWindow, kEventWindowHidden },
		{ kEventClassWindow, kEventWindowShown },

		{ kEventClassKeyboard, kEventRawKeyDown },
		{ kEventClassKeyboard, kEventRawKeyUp },
		{ kEventClassKeyboard, kEventRawKeyModifiersChanged },

		{ kEventClassMouse, kEventMouseDown },
		{ kEventClassMouse, kEventMouseUp },
		{ kEventClassMouse, kEventMouseMoved },
		{ kEventClassMouse, kEventMouseDragged },
		{ kEventClassMouse, kEventMouseWheelMoved },
	};
		
	EventTypeSpec appEvents[] =
	{
		{ kEventClassApplication, kEventAppActivated },
		{ kEventClassApplication, kEventAppDeactivated },
		{ kEventClassApplication, kEventAppHidden },
		{ kEventClassApplication, kEventAppShown },

		{ kEventClassCommand, kEventProcessCommand }
	};

	OSStatus err;
	IBNibRef theNib;

	// Load the menu bar from the nib
	err = CreateNibReference (CFSTR("main"), &theNib);  
	if (!err)
		SetMenuBarFromNib (theNib, CFSTR("MenuBar"));

	EnterMovies();

	//
	// Tweak the "About <game>" menu item to use the CFBundleName from our plist.
	//
	MenuRef appMenu = NULL;
	MenuItemIndex menuIndex;

	// Get the menu item and menu ref that correspond to the 'abou' command
	MenuRef rootMenu = AcquireRootMenu();
	err = GetIndMenuItemWithCommandID(rootMenu, 'abou', 1, &appMenu, &menuIndex);
	CFRelease(rootMenu); // supercedes ReleaseMenu() (which is now deprecated) starting with 10.2
	
	// Get the existing menu item string (e.g. "About ")
	CFStringRef menuString = NULL;
	err = CopyMenuItemTextAsCFString (appMenu, menuIndex, &menuString);

	// Get the real game name string, stored as EKSGameName in the plist.
	gAppName = (CFStringRef)CFBundleGetValueForInfoDictionaryKey(CFBundleGetMainBundle(), CFSTR("EKSGameName"));
	
	// Concatenate the 2 strings and set the menu item string to the resul.
	CFStringRef newString = CFStringCreateWithFormat(NULL, NULL, CFSTR("%@%@"), menuString, gAppName);
	err = SetMenuItemTextWithCFString (appMenu, menuIndex, newString);

	if (newString)	CFRelease (newString);
	if (menuString)	CFRelease (menuString);

	// Check the minimum OS version
	SInt32 majorVersion, minorVersion, bugFixVersion;

	if ((Gestalt(gestaltSystemVersionMajor, &majorVersion) == noErr) &&
		(Gestalt(gestaltSystemVersionMinor, &minorVersion) == noErr) &&
		(Gestalt(gestaltSystemVersionBugFix, &bugFixVersion) == noErr))
	{
		// as written, this should handle up to OS X 99.99.99
		gOSXVersion32 = ((majorVersion / 10) << 20) |
		((majorVersion % 10) << 16) |
		((minorVersion / 10) << 12) |
		((minorVersion % 10) << 8) |
		((bugFixVersion / 10) << 4) |
		((bugFixVersion % 10) << 0);
	}
	
	// We require 10.4.0 and later.
	if (gOSXVersion32 < 0x100400)
	{
		SystemMessage("This game requires Mac OS X 10.4.0 or later.");
		exit(0);
	}





#if BETA
{
	AlertStdCFStringAlertParamRec paramRec;
	DialogRef dialog;
	DialogItemIndex itemIndex;
	CFStringRef errorString = NULL;
	errorString = CFStringCreateWithCString (NULL, "Do you agree to not redistribute this program\nand to stop using it by December 1, 2008?", kCFStringEncodingMacRoman);
	CFStringRef defString = NULL;
	defString = CFStringCreateWithCString (NULL, "Yes, I agree!", kCFStringEncodingMacRoman);
	CFStringRef cancelString = NULL;
	cancelString = CFStringCreateWithCString (NULL, "No, I DON'T agree.", kCFStringEncodingMacRoman);
	paramRec.version = kStdCFStringAlertVersionOne;
	paramRec.movable = TRUE;
	paramRec.helpButton = FALSE;
	paramRec.defaultText = defString;
	paramRec.cancelText = cancelString;
	paramRec.otherText = NULL;
	paramRec.defaultButton = kAlertStdAlertOKButton;
	paramRec.cancelButton = kAlertStdAlertCancelButton;
	paramRec.position = kWindowDefaultPosition;
	paramRec.flags = 0;
inSM = TRUE;	
	CreateStandardAlert (kAlertNoteAlert, CFSTR("This is a beta test release.\nCopyright 2008 Everett Kaser\nAll rights reserved"), errorString, &paramRec, &dialog);
	RunStandardAlert (dialog, NULL, &itemIndex);
inSM = FALSE;

	CFRelease (errorString);
	CFRelease (defString);
	CFRelease (cancelString);

	if( itemIndex==kAlertStdAlertCancelButton ) {
		OtherError = TRUE;
//SystemMessage("Don't agree");
	}
}
#endif




	SetAppDataPath();

#if BETA
	bftime = time(NULL);
	bftm = localtime(&bftime);
	if( bftm->tm_year >= BETA_YEAR ) {
		if( bftm->tm_mon >= BETA_MONTH ) {
			if( bftm->tm_mday >= BETA_DAY ) {
//char tttmp[64];
				OtherError = TRUE;
//sprintf(tttmp, "Year %d:%d Month %d:%d Day %d:%d", bftm->tm_year, BETA_YEAR, bftm->tm_mon, BETA_MONTH, bftm->tm_mday, BETA_DAY);
//SystemMessage(tttmp);
			}
		}
	}
	
	Kfindclose();
	FirstNextList = new FindFileList();
	strcpy(TmpPath, "/tmp/*");
	Kff(FALSE, FALSE);
	// Point the list iterator to the first file entry
	FirstNextList->iter = FirstNextList->mList.begin();
	
	// Call Kfindnext to get the first item
	bfr = Kfindnext(&bfb[0], &bfa);
	while( !bfr ) {
		if( -1 != stat((const char *)&bfb[0], &fstats) ) {
			bftm = localtime(&fstats.st_mtime);
			if( bftm->tm_year >= BETA_YEAR ) {
				if( bftm->tm_mon >= BETA_MONTH ) {
					if( bftm->tm_mday >= BETA_DAY ) {
						OtherError = TRUE;
//SystemMessage("Bad file");
					}
				}
			}
		}
		if( FirstNextList ) 	{
			if( FirstNextList->iter != FirstNextList->mList.end() ) {
				// Found a match, return the data for this entry
				bfa = (*(FirstNextList->iter))->attributes;
				strcpy((char *)bfb, (*(FirstNextList->iter))->name.c_str());
				FirstNextList->iter++;
			} else bfr = 1;
		} else bfr = 1;
	}
	Kfindclose();
#endif

	// try to load the .KE file
	if( !LoadKEFile("main.ke", &mainKE) || OtherError )
		exit(0);

#if _DEBUG
	DWORD majorVers = GETMEMB(mainKE.pKode+KHDR_MAJOR_INTERP_REV);
	DWORD minorVers = GETMEMB(mainKE.pKode+KHDR_MINOR_INTERP_REV);
	printf ("KINT v%d.%d\n", INTERP_MAJOR_REV, INTERP_MINOR_REV);
	printf ("%s v%d.%d\n", CFStringGetCStringPtr(gAppName, kCFStringEncodingMacRoman), majorVers, minorVers);
#endif

	SysStartup();
	
	require( CreateMacWindow(), cantCreateWindow );

	InitScreen(0, 0);

	// call the BEGIN_STARTUP code in the KE module (tell it we're starting up and
	// about to create the main program window)
	KEmsg(&mainKE, KHDR_BEGIN_STARTUP, 0, 0, 0);

//	require( CreateMacWindow(), cantCreateWindow );

    InstallApplicationEventHandler(NewEventHandlerUPP(appCommandHandler), GetEventTypeCount(appEvents), appEvents, NULL, NULL);
	InstallWindowEventHandler (gMacWindow, GetwindowEventHandlerUPP(), GetEventTypeCount(windowEvents), windowEvents, NULL, NULL);

	// call the END_STARTUP code in the KE module (tell it to finish setting up, and
	// that the main program window has now been created.
	KEmsg(&mainKE, KHDR_END_STARTUP, 0, 0, 0);

	ShowWindow (gMacWindow);

	// send one IDLE message to the KASM code in case it needs to do anything once
	// right up front after everything else has been initialized.
	KEmsg(&mainKE, KHDR_IDLE, 0, 0, 0);

	ProgramReady = TRUE;

	// Install an idle timer that will fire IDLE_MS from now and repeat every 10IDLE_MS.
	InstallEventLoopIdleTimer(GetCurrentEventLoop(), (kEventDurationMillisecond * IDLE_MS), (kEventDurationMillisecond * IDLE_MS), NewEventLoopIdleTimerUPP(sIdleTimerProc), 0, &sIdleTimerRef);

	// Run the OS X event loop
	RunApplicationEventLoop();

cantCreateWindow:
	return 0;
}

//**********************************************************************
//* Called from PMACH.CPP when the .KE wants the program to terminate
//* Do whatever's necessary to make the program terminate.
//*
void KillProgram()
{
	KEmsg(&mainKE, KHDR_SHUT_DOWN, 0, 0, 0);
	SysCleanup();
	free(mainKE.pKode);
	StopTimer();
	exit(0);
}

//***********************************************************************
void KINT_SetWindowPos(DWORD r11, DWORD r12, DWORD r13, DWORD r14, DWORD r15)
{
#if _DEBUG
	printf("KINT_SetWindowPos, cmd: %08x, h: %d, w: %d, top: %d, left: %d\n", r11, r12, r13, r14, r15);
#endif

	BOOL t = InSetWindowPos;
	InSetWindowPos = TRUE;
	// We ignore the position coords and only resize. After that's done, we recenter.
	if (gMacWindow)
	{
		Rect windowBounds;
		(void)GetWindowBounds (gMacWindow, kWindowContentRgn, &windowBounds);
		
		windowBounds.right = windowBounds.left + r13;
		windowBounds.bottom = windowBounds.top + r12;

		(void)SetWindowBounds (gMacWindow, kWindowContentRgn, &windowBounds);
		RepositionWindow (gMacWindow, NULL, kWindowCenterOnMainScreen);
	}
	else
	{
#if _DEBUG
		Debugger();
#endif
	}
	InSetWindowPos = t;
}

//**************************************************************************
void KINT_GetWindowPos(DWORD *r11, DWORD *r12, DWORD *r13, DWORD *r14, DWORD *r15)
{
	if (!gMacWindow) return;
	
	Rect windowBounds;
	(void)GetWindowBounds (gMacWindow, kWindowContentRgn, &windowBounds);
	
//	*r15 = windowBounds.left; // x
//	*r14 = windowBounds.top; // y
	*r15 = 0; // x
	*r14 = 0; // y
	*r13 = windowBounds.right - windowBounds.left; // w
	*r12 = windowBounds.bottom - windowBounds.top; // h
	*r11 = 0;

#if _DEBUG
	printf("KINT_GetWindowPos, cmd: %08x, h: %d, w: %d, top: %d, left: %d\n", *r11, *r12, *r13, *r14, *r15);
#endif
}

static pascal void sTimerProc(EventLoopTimerRef inTimerRef, void *inUserData)
{
	if (!ProgramReady) return;

	// Add the KINT timer message, which will wake up the idle timer and process messages.
	AddMsg(KHDR_TIMER, 0, 0, 0);
}

static EventLoopTimerUPP sTimerUPP = NewEventLoopTimerUPP(sTimerProc);
//******************************************************************
void StartTimer(DWORD timval)
{
	// If the period has changed, kill off the existing event loop timer.
	if (timval != sTimerPeriod)
		StopTimer();
	
	// Install a repeating millisecond timer with our new interval.
	sTimerPeriod = timval;
	if (!sTimerRef)
		InstallEventLoopTimer(GetCurrentEventLoop(), (kEventDurationMillisecond * sTimerPeriod), (kEventDurationMillisecond * sTimerPeriod), sTimerUPP, 0, &sTimerRef);
}

//******************************************************************
void StopTimer()
{
	sTimerPeriod = 0;
	if (sTimerRef)
	{
		RemoveEventLoopTimer(sTimerRef);
		sTimerRef = NULL;
	}
}

//******************************************************************
void SetPhysicalScreenSize (void)
{
#if FAKE_SCREEN_SIZE
	int width = FAKE_SCREEN_SIZE_W;
	int height = FAKE_SCREEN_SIZE_H;
#else
	// Get the resolution of the main display
	CGRect mainRect = CGDisplayBounds (kCGDirectMainDisplay);
	
	int height = mainRect.size.height;
	int width = mainRect.size.width;
#endif
	
/*	
	// Compensate for the menubar
	height -= GetMBarHeight();

	PhysicalScreenSizeIndex = 0;
	PhysicalScreenW = 640;
	PhysicalScreenH = 480;

	// Try to find the largest windowSize that fits on this display
	int i = 0;
	do
	{
		if ((windowSizes[i].width <= width) && (windowSizes[i].height <= height))
		{
			PhysicalScreenSizeIndex = i;
			PhysicalScreenW = windowSizes[i].width;
			PhysicalScreenH = windowSizes[i].height;
		}
		i++;
	}
	while (windowSizes[i].width != 0);
*/
	PhysicalScreenSizeIndex = 0;
	PhysicalScreenW = width;
	PhysicalScreenH = height;

	// Try to find the largest windowSize that fits on this display
	for(PhysicalScreenSizeIndex=0; (windowSizes[PhysicalScreenSizeIndex].width <= width) && (windowSizes[PhysicalScreenSizeIndex].height <= height); PhysicalScreenSizeIndex++) ;
	if( PhysicalScreenSizeIndex ) --PhysicalScreenSizeIndex;
#if _DEBUG
	printf ("PhysW: %d PhysH: %d, PhysIndex: %d\n", PhysicalScreenW, PhysicalScreenH, PhysicalScreenSizeIndex);
#endif
}

//****************************************
bool CreateMacWindow (void)
{
	bool retVal = false;
	SetPhysicalScreenSize ();

	Rect windowBounds;
	windowBounds.top = 0;
	windowBounds.left = 0;
	windowBounds.right = windowSizes[PhysicalScreenSizeIndex].width;
	windowBounds.bottom =windowSizes[PhysicalScreenSizeIndex].height;
	WindowAttributes attribs = kWindowCloseBoxAttribute | kWindowCollapseBoxAttribute | kWindowStandardHandlerAttribute | kWindowResizableAttribute | kWindowFullZoomAttribute;

#if USE_HIVIEW
	attribs |= kWindowCompositingAttribute;
#endif

	require_noerr( CreateNewWindow (kDocumentWindowClass, attribs, &windowBounds, &gMacWindow), cantCreateWindow );
	require_noerr( CreateCGContextForPort(GetWindowPort(gMacWindow), &gCGContext), cantCreateCGContext );

#if USE_HIVIEW
	HIViewRef contentRef; contentRef = NULL;
	HIViewFindByID(HIViewGetRoot(gMacWindow), kHIViewWindowContentID,  &contentRef);
	static const EventTypeSpec myHIViewSpec = { kEventClassControl, kEventControlDraw };

	OSStatus err; err = InstallEventHandler(GetControlEventTarget (contentRef),
					NewEventHandlerUPP(drawEventHandler),
					GetEventTypeCount(myHIViewSpec),
					&myHIViewSpec,
					(void*) contentRef,
					NULL);
#endif

	retVal = true;
	
cantGetMainDevice:
cantCreateWindow:
cantCreateCGContext:
	return retVal;
}

//****************************************
void InitScreen(long minW, long minH)
{
#if _DEBUG
printf("InitScreen %d, %d\n", minW, minH);
#endif
	int			redraw;
	WORD		o, ow, oh;

	if( minW==-1 ) {
		ow = kBuf->logW;
		oh = kBuf->logH;
		minW = minH = 0;
	} else {
		ow = (WORD)minW;
		oh = (WORD)minH;
	}

	if( kBuf!=NULL ) {
		DeleteBMB(kBuf);
		redraw = 1;
	} else redraw = 0;

	SetPhysicalScreenSize();
	
	if( PhysicalScreenW > minW ) minW = PhysicalScreenW;
	if( PhysicalScreenH > minH ) minH = PhysicalScreenH;

	kBuf = CreateBMB((WORD)minW, (WORD)minH, TRUE);
	if( NULL==kBuf ) {
		SystemMessage("Failed creating screen BITMAP");
		exit(0);
	}

	// initialize the ENTIRE screen bitmap to black
	o = kBuf->orient;
	kBuf->orient = BMO_NORMAL;
	RectBMB(kBuf, 0, 0, kBuf->physW, kBuf->physH, 0x00000000, (DWORD)-1);
	kBuf->orient = o;

	kBuf->logW = ow;
	kBuf->logH = oh;

	if( kBuf->logW == 0 ) {
		kBuf->logW = PhysicalScreenW;
		kBuf->logH = PhysicalScreenH;
	}

	if( redraw ) {
		// call the SCREEN_CHANGE code in the KE module
		AddMsg(KHDR_SCREEN_CHANGE, 0, 0, 0);
		SendNextMsg();
	}
}

//********************************************************************
// Display the NULL-terminated message in 'pMem' using OS resources.
void SystemMessage(char *pMem)
{
	DialogRef dialog;
	DialogItemIndex itemIndex;

	CFStringRef errorString = NULL;
	require ( errorString = CFStringCreateWithCString (NULL, pMem, kCFStringEncodingMacRoman), cantCreateString );

inSM = TRUE;	
	CreateStandardAlert (kAlertStopAlert, CFSTR("This humble servant says:"), errorString, NULL, &dialog);
	RunStandardAlert (dialog, NULL, &itemIndex);
inSM = FALSE;
cantCreateString:
	if (errorString) CFRelease (errorString);
}

//*********************************************************************
// store the appropriate End-Of-Line sequence into buffer 'pMem' and
// return the number of bytes written to 'pMem'.
int GetSystemEOL(BYTE *pMem)
{
#if _DEBUG
	printf("GetSystemEOL\n");
#endif
	*pMem++ = 0x0a;
	return 1;
}

//***************************************
PBMB CreateBMB(WORD w, WORD h, WORD sysBM)
{
	PBMB	pX;

// enable to force crash to test if SYMBOLS have been properly stripped
	pX = (PBMB)malloc(sizeof(BMB));
	if( pX!=NULL )
	{
//		pX->linelen = 4*((kPixSize*w+kPixSize)/4);	// force line of pixels to take even number of DWORDS
		pX->linelen = 4*w;	// force line of pixels to take even number of DWORDS
//pX->linelen = 4*((3*w+3)/4);
		pX->pBits = NULL;
		pX->bitmapContext = NULL;
		if( !sysBM && NULL==(pX->pBits=(BYTE *)malloc(pX->linelen*h)) )
		{
			free(pX);
			pX = NULL;
		}
		else
		{
			pX->pAND = NULL;
			pX->pNext = (void *)pbFirst;
			if( pbFirst!=NULL ) pbFirst->pPrev = (void *)pX;
			pX->pPrev = NULL;
			pbFirst = pX;
			pX->physW = w;
			pX->physH = h;
			pX->clipX1 = pX->clipY1 = 0;
			pX->clipX2 = (int)w;
			pX->clipY2 = (int)h;
			pX->logW = w;
			pX->logH = h;
			pX->orient = BMO_NORMAL;
			if (sysBM)
			{
				pX->pBits=(BYTE *)malloc(pX->linelen*h);
				CGColorSpaceRef colorSpace = CGColorSpaceCreateWithName(kCGColorSpaceGenericRGB);
				pX->bitmapContext = CGBitmapContextCreate (pX->pBits, w, h, 8, pX->linelen, colorSpace, kCGImageAlphaNoneSkipFirst);
				CGColorSpaceRelease(colorSpace);
//				pX->pBits = (BYTE*)CGBitmapContextGetData (pX->bitmapContext);
				pX->cgImage = NULL;
			}
		}
	}
#if _DEBUG
printf("CreateBMB w: %d h: %d sys: %d, result: %08x\n", w, h, sysBM, pX);
#endif
	return pX;
}

//***************************************
void DeleteBMB(PBMB pX)
{
#if _DEBUG
printf("DeleteBMB %08x\n", pX);
#endif
	if( pX!=NULL )
	{
		if( pX->pAND!= NULL )
			free(pX->pAND);
		else if( pX->pBits != NULL )
			free(pX->pBits);
		if (pX->bitmapContext)
			CGContextRelease (pX->bitmapContext);
		if( pX==pbFirst )
		{
			pbFirst = (PBMB)pX->pNext;
			if( pbFirst!=NULL ) pbFirst->pPrev = NULL;
		}
		else
		{
			if( pX->pNext!=NULL ) ((PBMB)pX->pNext)->pPrev = pX->pPrev;
			if( pX->pPrev!=NULL ) ((PBMB)pX->pPrev)->pNext = pX->pNext;
		}
		free(pX);
	}
}

//**********************************************************************
void InvalidateScreen(int x1, int y1, int x2, int y2)
{
#if _DEBUG && DEBUG_UPDATES
printf("InvalidateScreen %d, %d to %d, %d (win: %08x)\n", x1, y1, x2, y2, gMacWindow);
#endif
	if (!gMacWindow) return;
	
	gUpdateImage = true;
#if USE_HIVIEW
	HIRect dirtyRect;
	dirtyRect.origin.x = x1;
	dirtyRect.origin.y = y1;
	dirtyRect.size.width = x2-x1;
	dirtyRect.size.height = y2-y1;
#else
	Rect dirtyRect;
	dirtyRect.left = x1;
	dirtyRect.right = x2;
	dirtyRect.top = y1;
	dirtyRect.bottom = y2;
#endif

#if USE_HIVIEW
	HIViewRef contentRef = NULL;
	HIViewFindByID(HIViewGetRoot(gMacWindow), kHIViewWindowContentID,  &contentRef);
	HIViewSetNeedsDisplayInRect(contentRef, &dirtyRect, true);
#else	
	InvalWindowRect (gMacWindow, &dirtyRect);
#endif
}

//**********************************************************************
void FlushScreen()
{
#if _DEBUG && DEBUG_UPDATES
printf("FlushScreen\n");
#endif
	HIWindowFlush ((HIWindowRef) gMacWindow);
}

#pragma mark -

//-----------------------------------------------------------------------------
// ConvertCFStringToUnicode
//-----------------------------------------------------------------------------
// Given a CFString, convert to a UniChar, suitable for FSRef file calls.

static int ConvertCFStringToUnicode (const CFStringRef inString, UniChar *outString)
{
	*outString = 0;
	
	assert(inString);
	assert(outString);
	if (!inString || !outString) return 0;
	
	CFIndex length = CFStringGetBytes (inString, CFRangeMake(0, CFStringGetLength(inString)), kCFStringEncodingUnicode, 0, false, (UInt8*)outString, 255, NULL);
	return length;
}

char BPname[PATHLEN];

void BuildPath(const char *pMem, BYTE inFolderType, BOOL AppData) {

	char	*p;
	bool	noExt = inFolderType & DIR_NOEXT;
	bool	isJpg = inFolderType & DIR_LOSSY; 
	
	// Replace all "\" with "/" in the incoming path
	strcpy(BPname, pMem);
	for(p=BPname; *p; p++) if( *p=='\\' ) *p = '/';
	
	if( AppData ) strcpy(TmpPath, AppDataPath);
	else strcpy(TmpPath, PgmPath);

	switch (inFolderType & 0x0f) {
	 case DIR_GAME:
		strcat(TmpPath, BPname);
		break;
	 case DIR_UI:
		strcat(TmpPath, "UI/");
		strcat(TmpPath, BPname);
		if( !noExt ) {
			if (isJpg) strcat(TmpPath, ".jpg");
			else strcat(TmpPath, ".bmp");		
		}
		break;
	 case DIR_IMAGES:
		strcat(TmpPath, "IMAGES/");
		strcat(TmpPath, BPname);
		if (!noExt) {
			if (isJpg) strcat(TmpPath, ".jpg");
			else strcat(TmpPath, ".bmp");		
		}
		break;
	 case DIR_BACKGNDS:
		strcat(TmpPath, "Backgnds/");
		strcat(TmpPath, BPname);
		if( !noExt ) {
			if (isJpg) strcat(TmpPath, ".jpg");
			else strcat(TmpPath, ".bmp");		
		}
		break;
	 case DIR_MUSIC:
		strcat(TmpPath, "Music/");
		strcat(TmpPath, BPname);
		if( !noExt ) strcat(TmpPath, ".mid");		
		break;
	 case DIR_SOUNDS:
		strcat(TmpPath, BPname);
		if( !noExt ) strcat(TmpPath, ".wav");		
		break;
	 case DIR_GAME_IMG:
		strcat(TmpPath, BPname);
		if( !noExt ) {
			if (isJpg) strcat(TmpPath, ".jpg");
			else strcat(TmpPath, ".bmp");		
		}
		break;
	 case DIR_UI_COMMON:
		strcat(TmpPath, "CommonUI/");
		strcat(TmpPath, BPname);
		if( !noExt ) {
			if (isJpg) strcat(TmpPath, ".jpg");
			else strcat(TmpPath, ".bmp");		
		}
		break;
	 default:
		break;
	}
}

//*********************************************************************************
// make a sub-folder of the DIR_GAME folder or a sub-folder thereof
// return non-zero if successful, else return zero if fail
int KMakeFolder(char *fname)
{
	int result;
	char	*s, *d;
	
#if _DEBUG && DEBUG_FILE
	printf ("KMakeFolder: %s\n", fname);
#endif
	s = fname;
	d = "../COMMONUI";
	for(; *s; s++, d++) if( toupper(*s) != *d ) break;
	
	if( !*s ) {
		strcpy(TmpPath, AppDataPath);
		strcat(TmpPath, "CommonUI");
	} else BuildPath(fname, DIR_GAME, TRUE);
	
	result = mkdir(TmpPath, 0777);

	// If the path already exists, don't return an error
	if( result != 0 && errno == EEXIST ) result = 0;
	
	return result;
}

//*********************************************************************************
// Delete an empty folder (can NOT contain ANY files)
void KDelFolder(char *fname) {

#if _DEBUG && DEBUG_FILE
//	MacNotImplemented();
	printf ("KDelFolder: %s\n", fname);
#endif

	BuildPath(fname, DIR_GAME, TRUE);
	rmdir(TmpPath);
}

//*********************************************************************************
// return file handle if successful, else -1
DWORD Kopen(char *fname, int oflags, BYTE dir) {

	// if O_WRBIN or O_WRBINNEW
		// if open(O_RDBIN) in AppData
			// close file
		// else if open(O_RDBIN) in PgmPath
			// close file
			// build AppData filename and copy to tmpbuf
			// KCopyFile(fname, tmpbuf);
		// open(oflags) in AppData
		// return handle
	// else
		// if open(O_RDBIN) in AppData
			// return handle
		// else if open(O_RDBIN) in PgmData
			// return handle
	// return -1

	char	*p, tmpbuf[PATHLEN];
	FILE	*fp;
	char	mode[8];

	switch (oflags) {
	 case O_WRBIN:
		strcpy(mode, "r+b");
		break;
	 case O_WRBINNEW:
		strcpy(mode, "w+b");
		break;
	 case O_RDBIN:
	 default:
		strcpy(mode, "rb");
		break;
	}

	if( oflags==O_RDBIN ) {	// if just READING
		BuildPath(fname, dir, TRUE);	// try AppData first
		fp = fopen(TmpPath, mode);
		if( fp==NULL ) {
			BuildPath(fname, dir, FALSE);	// try PgmPath second
			fp = fopen(TmpPath, mode);
		}
	} else {				// if WRITING
		BuildPath(fname, dir, TRUE);	// try AppData first
		strcpy(tmpbuf, TmpPath);		// save AppData target name
		fp = fopen(tmpbuf, "rb");
		if( fp==NULL ) {
			// make sure target AppData path exists
			for(p=tmpbuf+strlen(tmpbuf)-1; p>tmpbuf && *p!='/'; p--) ;	// find end of 'path' portion
			if( p > tmpbuf ) {
				char	*b[64];
				long	bcnt;

				bcnt = 0;
				b[bcnt] = p;
				*b[bcnt] = 0;	// cut off file name portion, just keep path portion
				++bcnt;
				while( -1==stat(tmpbuf, &fstats) ) {	// find 'lowest' portion of path that already exists
					for(b[bcnt]=b[bcnt-1]-1; b[bcnt]>tmpbuf && *b[bcnt]!='/'; --b[bcnt]) ;	// find next higher folder
					if( b[bcnt] <= tmpbuf ) break;
					*b[bcnt] = 0;
					++bcnt;
				}
				if( b[bcnt-1] > tmpbuf ) {
					while( bcnt>1 ) {			// create the rest of the path folders that didn't already exist
						--bcnt;
						*b[bcnt] = '/';
//						mkdir(tmpbuf);
					}
				}
				*b[0] = '/';
			}
			BuildPath(fname, dir, FALSE);	// try PgmPath second
			fp = fopen(TmpPath, "rb");
			if( fp != NULL ) {
				fclose(fp);
				KCopyFile(fname, tmpbuf);
			}
		} else fclose(fp);
		fp = fopen(tmpbuf, mode);
	}
	if( fp==NULL ) fp = (FILE *)-1;
	return (DWORD)fp;
}

//********************************************************************************
// return file length if successful, else -1
DWORD Kstat(char *fname, BYTE dir, DWORD *modtim, BYTE *filtyp) {

	FILE	*fp;
	DWORD	len;
	
	BuildPath(fname, dir, TRUE);	// try AppData first
	if( -1==stat(TmpPath, &fstats) ) {
		BuildPath(fname, dir, FALSE);
		if( -1==stat(TmpPath, &fstats) ) return 0xFFFFFFFF;
	}
	*modtim = fstats.st_mtime;
	if( (fstats.st_mode & S_IFMT)==S_IFDIR ) *filtyp = 2;
	else *filtyp = 1;
	
	fp = fopen(TmpPath, "rb");
	if( fp==NULL ) return 0xFFFFFFFF;
	
	fseek(fp, 0, SEEK_END);
	len = ftell(fp);
	fclose(fp);
	
	return len;
}

//********************************************************************************
// In KINT/KASM, 'origin' has these values (in case your language is different):
//		SEEK_SET    0
//		SEEK_CUR    1
//		SEEK_END    2
// return new file pointer position or -1 if error
DWORD Kseek(DWORD fhandle, DWORD foffset, int origin)
{
	DWORD retVal = -1;
	require (fhandle != (DWORD)-1, badHandle);
	require (fseek ((FILE*)fhandle, foffset, origin) == 0, cantSeek);
	
	retVal = ftell ((FILE*)fhandle);
	
#if _DEBUG && DEBUG_FILE
	printf("Kseek %08x offset: %d, origin: %02x retval: %08x\n", fhandle, foffset, origin, retVal);
#endif
	
badHandle:
cantSeek:
	return retVal;
}

//*********************************************************************************
void Kclose(DWORD fhandle)
{
#if _DEBUG && DEBUG_FILE
	printf("Kclose %08x\n", fhandle);
#endif
	
	require (fhandle != (DWORD)-1, badHandle);
	
	fclose ((FILE*)fhandle);
	
badHandle:
	return;
}

//********************************************************************************
//	return 1 if at EOF, 0 if not at EOF, -1 if error
DWORD Keof(DWORD fhandle)
{
	DWORD retVal = -1;
	require (fhandle != (DWORD)-1, badHandle);
	
	retVal = feof ((FILE*)fhandle);
	if (retVal)
		retVal = 1;
	
#if _DEBUG && DEBUG_FILE
	printf("Keof %08x, %d\n", fhandle, retVal);
#endif
	
badHandle:
	return retVal;
}

//******************************************************
DWORD Kread(DWORD fhandle, BYTE *pMem, DWORD numbytes)
{
	DWORD retVal = 0;
	require (fhandle != (DWORD)-1, badHandle);
	require (pMem, badPointer);
	
	retVal = fread (pMem, 1, numbytes, (FILE*)fhandle);
	
#if _DEBUG && DEBUG_FILE
	printf("Kread %08x, numbytes: %d, result: %d\n", fhandle, numbytes, retVal);
#endif
	
badHandle:
badPointer:	
	return retVal;
}

//*******************************************************
DWORD Kwrite(DWORD fhandle, BYTE *pMem, DWORD numbytes)
{
	DWORD retVal = 0;
	require (fhandle != (DWORD)-1, badHandle);
	require (pMem, badPointer);
	
	retVal = fwrite (pMem, 1, numbytes, (FILE*)fhandle);
	
#if _DEBUG && DEBUG_FILE
	printf("Kwrite %08x, numbytes: %d, result: %d\n", fhandle, numbytes, retVal);
#endif
	
badHandle:
badPointer:	
	return retVal;
}

#pragma mark -

//------------------------------------------------------------------------------------
// • wildicmp
//------------------------------------------------------------------------------------
// Case-insensitive wildcard matching.
// Matches against ? and *. Returns nonzero if match found, 0 if no match.

int wildicmp(const char *string, const char *wild)
{
	const char *charP = NULL, *wildP = NULL;

	// Attempt to match the string up to the first "*"
	while ((*string) && (*wild != '*'))
	{
		// No exact or ? match, so fail
		if ((*wild != '?') && (tolower(*wild) != tolower(*string)))
			return 0;

		// Otherwise move to the next char in each string
		wild++;
		string++;
	}

	// Match the rest of the string.
	while (*string)
	{
		// See if we hit a * token
		if (*wild == '*')
		{
			// If we're at the end of the wild string, we're done and we match.
			if (!*++wild)
				return 1;

			// wild and wildP are now the char after the * token
			wildP = wild;
			// advance to the next source char after this.
			charP = string+1;
		}
		// No * token, check for exact and ? matches.
		else if ((*wild == '?') || (tolower(*wild) == tolower(*string)))
		{
			wild++;
			string++;
		}
		else
		{
			// No match. If wildP is not NULL, we've got a * search and can consume another entry
			// in "string".
			// Otherwise we'll fail the match next time through the loop due to wild being NULL.
			wild = wildP;
			string = charP++;
		}
	}

	// Consume any * tokens at the end of the string
	while (*wild == '*')
		wild++;

	// If we're past the last * token in the wild string, we'll either be NULL or we won't
	// match (i.e. the wild string was longer than the source).
	return !*wild;
}

int stricmp(const char *s, const char *d) {

	while( *s && *d ) {
		if( tolower(*s) != tolower(*d) ) break;
		++s;
		++d;
	}
	if( *s < *d ) return -1;
	if( *s > *d ) return  1;
	return 0;
}

void Kff(BOOL appdata, BOOL removeExtension) {

	FSIterator		iter = NULL;
	FSCatalogInfo	*catInfos = NULL;
	HFSUniStr255	*names = NULL;
	const ItemCount	kMaxInfo = 64;
	string			fileName;
	OSStatus		err;
	FSRef			fsRef;
	bool			wild = false;
	char			macPath[1024], macName[256];
	char			*fp;
	const char		*p;

#if _DEBUG && DEBUG_FILE
	printf ("find first: %s\n", fileName.c_str());
#endif

	fileName = TmpPath;

	// Make sure we're using Mac-like file names
	strcpy (macPath, fileName.c_str());
	
	// If no wildcards are specified we can optimize
	for(p = macPath; *p && !wild; p++) if( *p == '*' || *p == '?' ) wild = true;
	if( !wild ) require_noerr(FSPathMakeRef((UInt8 *)macPath, &fsRef, NULL), cantMakePathRef);
	else {
		// Split the directory name and filename
		fp = strrchr(macPath, '/');
		if( fp ) {
			strcpy(macName, fp+1);
			*(fp+1) = '\0';
		} else {
			strcpy(macName, macPath);
			strcpy(macPath, "./");
		}
		
		// Get an FSRef for the directory
		Boolean isDirectory;
		err = FSPathMakeRef((UInt8 *)macPath, &fsRef, &isDirectory);
		if( err != noErr || !isDirectory ) return;
	}

	// Create a catalog iterator for the directory
	if( wild ) require_noerr( FSOpenIterator(&fsRef, kFSIterateFlat, &iter), cantOpenIterator);

	// Allocate some space to receive catalog info
	catInfos = new FSCatalogInfo[kMaxInfo];
	names = new HFSUniStr255[kMaxInfo];

	// Iterate through all catalog info for items in this directory
	do {
		// Grab a chunk of catalog info
		ItemCount actualItems = 0;
		ItemCount index = 0;
		FSCatalogInfoBitmap infoBits = kFSCatInfoNodeFlags + kFSCatInfoAllDates + kFSCatInfoFinderInfo + kFSCatInfoDataSizes;
		if( wild ) err = FSGetCatalogInfoBulk(iter, kMaxInfo, &actualItems, NULL, infoBits, catInfos, NULL, NULL, names);
		else {
			err = FSGetCatalogInfo(&fsRef, infoBits, &catInfos[0], &names[0], NULL, NULL);
			if( noErr == err ) {
				err = errFSNoMoreItems;
				actualItems = 1;
			}
		}
		while( (noErr == err || errFSNoMoreItems == err)  &&  actualItems > 0 ) {
			// Search for all files matching our search criteria
 			const HFSUniStr255 *curUniName = &names[index];
			const FSCatalogInfo *curInfo = &catInfos[index];
			
			// Convert the HFS unicode name to a MacRoman C string
			CFStringRef nameRef = CFStringCreateWithCharacters(NULL, curUniName->unicode, curUniName->length);
			if( nameRef ) {
				char curName[256];
				
				if( CFStringGetCString(nameRef, curName, sizeof(curName), kCFStringEncodingMacRoman) ) {
					if( wild == false || wildicmp(curName, macName) ) {
						UInt32 attribs = 0;
						
						if( curInfo->nodeFlags & kFSNodeIsDirectoryMask ) attribs |= 0x10;
						if( removeExtension ) {
							char *dotPos;
							
							if( dotPos = strrchr(curName, '.') ) *dotPos = 0;
						}
						// if already in list (duplicate in AppDataPath and PgmPath) skip it
						for(FirstNextList->iter = FirstNextList->mList.begin(); FirstNextList->iter != FirstNextList->mList.end(); FirstNextList->iter++) {
							if( !stricmp(curName, (*(FirstNextList->iter))->name.c_str()) ) break;
						}
						if( FirstNextList->mList.begin()==FirstNextList->mList.end() || FirstNextList->iter == FirstNextList->mList.end() ) {
							// Create an entry for this file for the linked list
							FileEntry *entry = new FileEntry(curName, attribs, appdata);
							FirstNextList->mList.push_back(entry);
						}
					}
				}
				CFRelease(nameRef);
			}
			--actualItems; // done with this item, process next
			++index;
		}
		// done with this chunk of catalog info, get next
	} while( err == noErr );
	
cantMakePathRef:
cantOpenIterator:
	if( iter )		FSCloseIterator(iter);
	if( catInfos )	delete[] catInfos;
	if( names )		delete[] names;
}

//******************************************************************
// Search for a filename (possibly containing '*' as wild card)
DWORD Kfindfirst(char *fname, BYTE *fbuf, BYTE *attrib, BYTE dir) {

	DWORD	retVal = (DWORD)-1;
	bool	removeExtension;
	
	Kfindclose();
	
	removeExtension = (dir != DIR_GAME && !(dir & DIR_NOEXT));

	// Create the linked list
	FirstNextList = new FindFileList();
	
	if( CopyAppData ) {
		BuildPath (fname, dir, TRUE);
		Kff(TRUE, removeExtension);
	}
	BuildPath(fname, dir, FALSE);
	Kff(FALSE, removeExtension);
	
	// Point the list iterator to the first file entry
	FirstNextList->iter = FirstNextList->mList.begin();
	
	// Call Kfindnext to get the first item
	if( retVal = Kfindnext(fbuf, attrib) ) Kfindclose();

	return retVal;
}

//********************************************************************
// Search for next filename in sequence started by Kfindfirst()
DWORD Kfindnext(BYTE *fbuf, BYTE *attrib) {

	if( FirstNextList ) 	{
		if( FirstNextList->iter != FirstNextList->mList.end() ) {
			// Found a match, return the data for this entry
			*attrib = (*(FirstNextList->iter))->attributes;
			strcpy((char *)fbuf, (*(FirstNextList->iter))->name.c_str());
			FirstNextList->iter++;
			return 0;
		}
	}
	return 0xFFFFFFFF;
}

//****************************************************
// Terminate the Kfindfirst()/Kfindnext() process
void Kfindclose() {

	if( FirstNextList != NULL ) {
		// Dispose of our list header
		delete FirstNextList;
		FirstNextList = NULL;
	}
}

#pragma mark -

//*************************************************
void Kdelete(char *fname, BYTE dir)
{
#if _DEBUG && DEBUG_FILE
	printf ("Kdelete: %s\n", fname);
#endif
	string fileName;

	require(fname, badFileName);

	BuildPath (fname, dir, TRUE);
	fileName = TmpPath;
	remove (fileName.c_str());

badFileName:
	return;
}

//*****************************************************
WORD KCopyFile(char *src, char *dst)
{
#if _DEBUG && DEBUG_FILE
	MacNotImplemented();
	printf ("KCopyFile: %s to %s\n", src, dst);
#endif
	return FALSE;
}

//*****************************************************
void KGetTime(KTIME *pT)
{
	CFTimeZoneRef timeZone = CFTimeZoneCopyDefault();
	CFGregorianDate curtime = CFAbsoluteTimeGetGregorianDate(CFAbsoluteTimeGetCurrent(), timeZone);
	
	if (timeZone)
		CFRelease(timeZone);

	pT->y = curtime.year;
	pT->mo= curtime.month;
	pT->d = curtime.day;
	pT->h = curtime.hour;
	pT->mi= curtime.minute;
	pT->s = curtime.second;
}

//******************************************************
void KTimeToString(BYTE *pMem)
{
	CFTimeZoneRef timeZone = CFTimeZoneCopyDefault();
	CFGregorianDate curtime = CFAbsoluteTimeGetGregorianDate(CFAbsoluteTimeGetCurrent(), timeZone);
	
	if (timeZone)
		CFRelease(timeZone);

	sprintf((char *)pMem, "%d-%d-%d %2.2d:%2.2d:%2.2d", curtime.month, curtime.day, curtime.year, curtime.hour, curtime.minute, curtime.second);
}

//****************************************************************************
//		Return a millisecond count which doesn't necessarily bear
//		any resemblance to "real" time, but rather is a counter that is
//		a reasonably accurate "millisecond counter" that may eventually
//		roll over to 0, but which should not do so more often than once a day,
//		and who's count value continues upwards (until such time as a roll-over
//		might occur).  This is accomplished on the MS-Windows platform using
//		the GetTickCount() function.  Other platforms may choose to implement
//		this in whatever way seems easiest, while meeting the above criteria.
//		This counter value is only used for timing the speed of animations
//		and brief delays in actions.  So, reasonable accuracy down to 5 or 10
//		milliseconds would be very nice, and the roll-over (if and when it
//		occurs) should only cause (at most) a slight burp in any animation
//		or screen action that is going on at the moment (but from which it
//		should recover almost immediately).
//    NOTE: It is IMPORTANT that his count be relative accurate, as it is
//		used for generating the elapsed time count for solving puzzles!
//	It could be done getting the time of day, and subtracting that from the
//  time of day when the program started, but the Windows GetTickCount() function
//	works nicely, too.

DWORD KGetTicks()
{
	static double sMillisecondsPerBeat = 0.0;
	
	// If this is the first time here, create a conversion factor
	// so we can convert the timebase registers to milliseconds.
	if (sMillisecondsPerBeat == 0.0)
	{
		mach_timebase_info_data_t timebase; 
		mach_timebase_info (&timebase);
		
		// Convert nanosecond result to milliseconds
		sMillisecondsPerBeat = (double) timebase.numer / (double) timebase.denom * 0.000001; 
	}

	// Read the timebase registers and convert to ms.
	uint64_t ticks = mach_absolute_time();
	return (DWORD)(ticks * sMillisecondsPerBeat);
}

void KSleep(DWORD tm)
{
	usleep(tm*1000);
}

//**********************************************************
void KSetCursor(WORD curtype)
{
//	curtype = 0 for normal, 1 for wait, 2 for hotlink
	ThemeCursor newCursor;
	switch (curtype)
	{
		case 0:		newCursor = kThemeArrowCursor; break;
		case 1:		newCursor = kThemeWatchCursor; break;
		case 2:		newCursor = kThemePointingHandCursor; break;
		default:	newCursor = kThemeArrowCursor; break;
	}
	SetThemeCursor (newCursor);
}

void KCaptureMouse(BOOL cap)
{
//	MacNotImplemented();
}

//**********************************************************
void KSetTitle(BYTE *pMem)
{
	require (pMem, badParam);
	require (gMacWindow, badParam);
	
	CFStringRef windowTitle; windowTitle = NULL;
	windowTitle = CFStringCreateWithCString (kCFAllocatorDefault, (const char *)pMem, kCFStringEncodingUTF8);
	require (windowTitle, cantCreateString);
	
	require_noerr( SetWindowTitleWithCFString (gMacWindow, windowTitle), cantSetWindowTitle );

badParam:
cantCreateString:
cantSetWindowTitle:
	if (windowTitle)
		CFRelease (windowTitle);
	return;
}

#pragma mark -

// TODO: line and BMB positioning is kinda hacky owing to accounting for font metrics.
// Instead of compensating by the margin.top/2, need to compensate by the height of the current font.

PMPrintSession printSession = NULL;
PMPrintSettings printSettings = NULL;
PMPageFormat pageFormat = NULL;
PMOrientation orientation = kPMPortrait;
PMPaper paper;
PMPaperMargins paperMargins;
double pageWidth, pageHeight;

bool printFontFixed = true;
CGFloat printFontSize = 12.0f;
bool printFontBold = false;
bool printFontItalic = false;

CGContextRef cgPrintingContext = NULL;

//***********************************************************
// Return 0 if fail, non-0 if succeed (and set pw, ph, Xppi, Yppi)
// Get everything setup for printing which is to follow
DWORD KPrintStart(DWORD *pw, DWORD *ph, DWORD *Xppi, DWORD *Yppi)
{
#if _DEBUG && DEBUG_PRINT
	printf("KPrintStart\n");
#endif

	DWORD retVal = 0;

	gInPrintLoop = true;

	require_noerr( PMCreateSession (&printSession), cantCreateSession);
	require_noerr( PMCreatePrintSettings (&printSettings), cantCreateSettings);
	require_noerr( PMSessionDefaultPrintSettings (printSession, printSettings), cantSetDefaults);

//	require_noerr( PMPrintSettingsSetJobName (printSettings, windowTitleRef), cantSetJobName);

	require_noerr( PMCreatePageFormat (&pageFormat), cantCreatePageFormat);
	require_noerr( PMSessionDefaultPageFormat (printSession,  pageFormat), cantGetDefaultPageFormat);

	require_noerr( PMSessionSetDestination (printSession, printSettings, kPMDestinationPreview, kPMDocumentFormatPDF, NULL), cantSetDestination);

	Boolean result;
	require_noerr( PMSessionValidatePageFormat (printSession, pageFormat, &result), cantValidateFormat);

	require_noerr( PMGetPageFormatPaper (pageFormat, &paper), cantGetPaper);
	require_noerr( PMPaperGetWidth (paper, &pageWidth), cantGetPaperWidth);
	require_noerr( PMPaperGetHeight (paper, &pageHeight), cantGetPaperHeight);
	require_noerr( PMPaperGetMargins (paper, &paperMargins), cantGetPaperMargins);

	require_noerr( PMGetOrientation (pageFormat, &orientation), cantGetPaperOrientation);
	// Landscape paper sizes have the width and height reversed.
	if (orientation == kPMLandscape || orientation == kPMReverseLandscape)
	{
		double temp;
		temp = pageHeight;
		pageHeight = pageWidth;
		pageWidth = temp;
	}
	*pw = pageWidth - (paperMargins.left + paperMargins.right);
	*ph = pageHeight - (paperMargins.top + paperMargins.bottom);

	// This may tickle the idle timer, so we have to set gInPrintLoop beforehand.
	require_noerr( PMSessionBeginCGDocument (printSession, printSettings, pageFormat), cantBeginSession);

	*Xppi = (DWORD)72;
	*Yppi = (DWORD)72;

#if _DEBUG && DEBUG_PRINT
printf("  w: %d h: %d x ppi: %d y ppi: %d\n", *pw, *ph, *Xppi, *Yppi);
#endif

	retVal = 1;
	
cantCreateSession:
cantCreateSettings:
cantSetDefaults:
cantSetJobName:
cantCreatePageFormat:
cantGetDefaultPageFormat:
cantSetDestination:
cantValidateFormat:
cantGetPaper:
cantGetPaperWidth:
cantGetPaperHeight:
cantGetPaperOrientation:
cantGetPaperMargins:
cantBeginSession:

	if (retVal == 0)
		KPrintStop();

	return retVal;
}

//*******************************************************
// Shutdown and clean up after printing, we're all through
void KPrintStop()
{
#if _DEBUG && DEBUG_PRINT
printf("KPrintStop\n");
#endif
	if (pageFormat)
	{
		PMRelease (pageFormat);
		pageFormat = NULL;
	}
	if (printSettings)
	{
		PMRelease (printSettings);
		printSettings = NULL;
	}
	if (printSession)
	{
		PMSessionEndDocument (printSession);
		PMRelease (printSession);
		printSession = NULL;
	}

	gInPrintLoop = false;
}

//*****************************************************
//	return non-0 if successful, 0 if fail
BYTE KPrintBeginPage()
{
#if _DEBUG && DEBUG_PRINT
printf("KPrintBeginPage\n");
#endif
	BYTE retVal = 0;

	require_noerr( PMSessionBeginPage(printSession, pageFormat, NULL), cantBeginPage);

	require_noerr( PMSessionGetCGGraphicsContext(printSession, &cgPrintingContext), cantGetContext);
	
	CGContextSelectFont(cgPrintingContext, printFontFixed ? "Courier" : "Helvetica", printFontSize, kCGEncodingMacRoman);
	retVal = 1;

cantBeginPage:
cantGetContext:

	return retVal;
}

//*****************************************************
//	return non-0 if successful, 0 if fail

BYTE KPrintEndPage()
{
#if _DEBUG && DEBUG_PRINT
printf("KPrintEndPage\n");
#endif
	if (PMSessionEndPage (printSession))
		return 0;
	else
		return 1;
}

//******************************************************
void KPrintText(int x, int y, char *str, DWORD clr)
{
#if _DEBUG && DEBUG_PRINT
printf("KPrintText x: %d y: %d %s\n", x, y, str);
#endif
	// TODO: deal with clr?
	CGContextShowTextAtPoint(cgPrintingContext, x + paperMargins.left, pageHeight - y - paperMargins.top, str, strlen(str));
}

//***********************************************************
void KPrintGetTextSize(char *str, DWORD *dw, DWORD *dh, DWORD *dasc)
{
#if _DEBUG && DEBUG_PRINT
printf("KPrintGetTextSize\n");
#endif
	CGPoint start, end;
	CGContextSetTextDrawingMode(cgPrintingContext, kCGTextInvisible);
	start = CGContextGetTextPosition(cgPrintingContext);
	CGContextShowText(cgPrintingContext, str, strlen(str));
	end = CGContextGetTextPosition(cgPrintingContext);
	CGContextSetTextDrawingMode(cgPrintingContext, kCGTextFill);
	CGContextSetTextPosition(cgPrintingContext, start.x, start.y);
	
	*dw = end.x - start.x;
	// TODO: deal with metrics for CG
	*dh = 12;
	*dasc = 10;
}

//**************************************************************************
void KPrintBMB(int xd, int yd, int wd, int hd, PBMB bmS, int xs, int ys, int ws, int hs)
{
#if _DEBUG && DEBUG_PRINT
printf("KPrintBMB\n");
#endif

	// Sherlock will pass in a NULL bmS if you try to print a puzzle, so we check for that and
	// exit early to avoid a crash.
	if (bmS == NULL) return;

	CGContextRef bitmapContext = NULL;
	CGColorSpaceRef colorSpace = CGColorSpaceCreateWithName(kCGColorSpaceGenericRGB);
	bitmapContext = CGBitmapContextCreate (bmS->pBits + (xs*kPixSize) + (bmS->linelen * ys), ws, hs, 8, bmS->linelen, colorSpace, kCGImageAlphaNoneSkipFirst);
	CGColorSpaceRelease (colorSpace);

	CGImageRef image = CGBitmapContextCreateImage (bitmapContext);

	CGRect cgRect = CGRectMake(xd + paperMargins.left, pageHeight - yd - hd - (paperMargins.top / 2), wd, hd);
	CGContextDrawImage (cgPrintingContext, cgRect, image);
	CGImageRelease (image);
	CGContextRelease (bitmapContext);
}

//**************************************************************************
void KPrintRect(int xd, int yd, int wd, int hd, DWORD fillcolor, DWORD edgecolor)
{
#if _DEBUG && DEBUG_PRINT
printf("KPrintRect\n");
#endif
	CGContextSaveGState(cgPrintingContext);

	CGColorSpaceRef cgColorSpace = CGColorSpaceCreateWithName (kCGColorSpaceGenericRGB);
	
	// fillcolor and edgecolor should be in ARGB format, so need to pack them into the float array in RGBA format.
	CGFloat tempColor[4];
	
	CGRect cgRect = CGRectMake(xd + paperMargins.left, pageHeight - yd - hd - (paperMargins.top / 2), wd, hd);

	if (fillcolor != 0xffffffff)
	{
		tempColor[0] = (float)((fillcolor & 0x00ff0000) >> 16) / 255.0f;	// R
		tempColor[1] = (float)((fillcolor & 0x0000ff00) >> 8) / 255.0f;		// G
		tempColor[2] = (float)(fillcolor & 0x000000ff) / 255.0f;			// B
		tempColor[3] = (float)((fillcolor & 0xff000000) >> 24) / 255.0f;	// A

		// The games use 0 to indicate an opaque alpha, so invert it for CoreGraphics
		tempColor[3] = 1.0f - tempColor[3];

		CGColorRef cgFillColor = CGColorCreate (cgColorSpace, tempColor);

		CGContextSetFillColorWithColor (cgPrintingContext, cgFillColor);
		CGContextFillRect (cgPrintingContext, cgRect);
		CGColorRelease (cgFillColor);
	}
	
	if (edgecolor != 0xffffffff)
	{
		tempColor[0] = (float)((edgecolor & 0x00ff0000) >> 16) / 255.0f;	// R
		tempColor[1] = (float)((edgecolor & 0x0000ff00) >> 8) / 255.0f;		// G
		tempColor[2] = (float)(edgecolor & 0x000000ff) / 255.0f;			// B
		tempColor[3] = (float)((edgecolor & 0xff000000) >> 24) / 255.0f;	// A
		
		// The games use 0 to indicate an opaque alpha, so invert it for CoreGraphics
		tempColor[3] = 1.0f - tempColor[3];

		CGColorRef cgStrokeColor = CGColorCreate (cgColorSpace, tempColor);

		CGContextSetStrokeColorWithColor (cgPrintingContext, cgStrokeColor);
		CGContextStrokeRect (cgPrintingContext, cgRect);
		CGColorRelease (cgStrokeColor);
	}
	CGColorSpaceRelease (cgColorSpace);

	CGContextRestoreGState(cgPrintingContext);
}

//************************************************************************************
void KPrintFont(WORD hlim, WORD flags, WORD *h, WORD *asc)
{
#if _DEBUG && DEBUG_PRINT
printf("KPrintFont\n");
#endif
	// flags:
	// bit 0: 1 = fixed-width (Win32: "Courier New" vs. "Tahoma")
	// bit 1: 1 = bold
	// bit 2: 1 = italic

	printFontFixed = (flags & 0x01);
	printFontSize = hlim - 2;
	printFontBold = (flags & 0x02);
	printFontItalic = (flags & 0x04);

	CGContextSelectFont(cgPrintingContext, printFontFixed ? "Courier" : "Helvetica", printFontSize, kCGEncodingMacRoman);
	// TODO: deal with bold/italic for CG
	// TODO: deal with metrics for CG
	*h = hlim;
	*asc = hlim - 2;
}

//*********************************************************
void KPrintLines(DWORD *pMem, WORD numpoints, DWORD clr)
{
#if _DEBUG && DEBUG_PRINT
printf("KPrintLines\n");
#endif

	int		i, j;

	CGContextSaveGState(cgPrintingContext);

	CGColorSpaceRef cgColorSpace = CGColorSpaceCreateWithName (kCGColorSpaceGenericRGB);
	
	// fillcolor and edgecolor should be in ARGB format, so need to pack them into the float array in RGBA format.
	CGFloat tempColor[4];
	
	tempColor[0] = (float)((clr & 0x00ff0000) >> 16) / 255.0f;	// R
	tempColor[1] = (float)((clr & 0x0000ff00) >> 8) / 255.0f;	// G
	tempColor[2] = (float)(clr & 0x000000ff) / 255.0f;			// B
	tempColor[3] = (float)((clr & 0xff000000) >> 24) / 255.0f;	// A
	
	// The games use 0 to indicate an opaque alpha, so invert it for CoreGraphics
	tempColor[3] = 1.0f - tempColor[3];

	CGColorRef cgStrokeColor = CGColorCreate (cgColorSpace, tempColor);
	CGContextSetStrokeColorWithColor (cgPrintingContext, cgStrokeColor);

	// KINT uses the first and last points once, and in between points twice,
	// where OS-X expects two points for every line segment, so we have to adjust.
	CGPoint *points = new CGPoint[2*(numpoints-1)];
	
	i = j = 0;
	points[i].x = EndianU32_LtoN(pMem[j++]) + paperMargins.left;
	points[i].y = pageHeight - EndianU32_LtoN(pMem[j++]) - (paperMargins.top / 2) - 2;
	
	for(++i; i < 2*(numpoints-1)-1; i+=2) {
		points[i].x = points[i+1].x = EndianU32_LtoN(pMem[j]) + paperMargins.left;
		j++;
		points[i].y = points[i+1].y = pageHeight - EndianU32_LtoN(pMem[j]) - (paperMargins.top / 2) - 2;
		j++;
	}
	points[i].x = EndianU32_LtoN(pMem[j++]) + paperMargins.left;
	points[i].y = pageHeight - EndianU32_LtoN(pMem[j++]) - (paperMargins.top / 2) - 2;
	
	CGContextStrokeLineSegments(cgPrintingContext, points, 2*(numpoints-1));
	
	delete[] points;

	CGColorRelease (cgStrokeColor);
	CGColorSpaceRelease (cgColorSpace);
	CGContextRestoreGState (cgPrintingContext);
}

#pragma mark -
#pragma mark • Audio

static QTCallBackUPP	sQTCallBackUPP;
static movieSound_t		*sMusicSound;

pascal void loopCallback (QTCallBack cb, long refCon)
{
	movieSound_t *qtSound = (movieSound_t *) refCon;

	if (qtSound->mMovie)
	{
		if (qtSound->mIsMusic)
		{
			MusicPlaying = 0;
			AddMsg(KHDR_MUSIC_DONE, 0, 0, 0);
		}
		if (qtSound->mLooping)
		{
			GoToBeginningOfMovie (qtSound->mMovie);
			StartMovie (qtSound->mMovie);
			CallMeWhen (qtSound->qtCallback, sQTCallBackUPP, (long) qtSound, triggerAtStop, 0, 0);
		}
		else
		{
			qtSound->mPlaying = false;
			if (qtSound->mAutoDelete)
			{
				// Add to orphan queue for deletion in the idle timer
				sSoundOrphans.push_back(qtSound);
			}
		}
	}
}

void SetupLoopCallback(movieSound_t *inSound)
{
	if (!inSound && !inSound->mMovie) return;

	if (!sQTCallBackUPP)
		sQTCallBackUPP = NewQTCallBackUPP (loopCallback);
	
	if (!inSound->qtCallback)
		inSound->qtCallback = NewCallBack (GetMovieTimeBase (inSound->mMovie), callBackAtExtremes + callBackAtInterrupt);
	(void)CallMeWhen (inSound->qtCallback, sQTCallBackUPP, (long) inSound, triggerAtStop, 0, 0);
}

movieSound_t * PlayFile_QT (const char *inMacPath, bool inLooping, bool inDisposeWhenDone, bool inMusic)
{
	OSType dataRefType;
	FSRef fsRef;
	movieSound_t *newSound = NULL;

	require_noerr(FSPathMakeRef((UInt8 *)inMacPath, &fsRef, NULL), cantCreatePath);

	newSound = (movieSound_t *)calloc (sizeof(movieSound_t), 1);
	newSound->mLooping = inLooping;
	newSound->mAutoDelete = inDisposeWhenDone;
	newSound->mIsMusic = inMusic;

	// Open the movie
	require_noerr( QTNewDataReferenceFromFSRef(&fsRef, 0, &(newSound->mDataRef), &dataRefType), cantCreateDataRef);
	require_noerr( NewMovieFromDataRef(&(newSound->mMovie), newMovieDontAskUnresolvedDataRefs, NULL, newSound->mDataRef, dataRefType), cantCreateMovie);

	// This is an audio-only movie, so disassociate any visual contexts from it
	SetMovieVisualContext (newSound->mMovie, NULL);
	
	// Go to the beginning of the movie
	SetMovieActive(newSound->mMovie, true);
	GoToBeginningOfMovie(newSound->mMovie);
	PrerollMovie(newSound->mMovie, 0, 0);

	SetupLoopCallback(newSound);
	
	// Start playback
	StartMovie(newSound->mMovie);
	newSound->mPlaying = true;

cantCreatePath:
cantCreateDataRef:
cantCreateMovie:
	if (newSound && newSound->mDataRef)
	{
		DisposeHandle(newSound->mDataRef);
		newSound->mDataRef = NULL;
	}
	return newSound;
}

void StopSound_QT (movieSound_t *inSound)
{
	if (!inSound) return;
	
	if (inSound->mMovie)
	{
		if (inSound->qtCallback)
		{
			CancelCallBack (inSound->qtCallback);
			DisposeCallBack (inSound->qtCallback);
			inSound->qtCallback = NULL;
		}
		
		if (inSound->mPlaying)
		{
			StopMovie(inSound->mMovie);
			SetMovieActive(inSound->mMovie, false);
			GoToEndOfMovie(inSound->mMovie);
			inSound->mPlaying = false;
		}
		
		DisposeMovie (inSound->mMovie);
		if (inSound->mDataRef)
			DisposeHandle (inSound->mDataRef);
			
		free (inSound);
	}	
}

//
// LBO - the following is ripped from Apple sample code:
// <http://developer.apple.com/technotes/tn/tn1195.html>
//

//////////
//
// PtrDataRef_AddFileNameExtension
//
// Tell the data handler to set
// the file name extension in the
// data reference.
//
//////////

static OSStatus PtrDataRef_AddFileNameExtension(
        ComponentInstance dataRefHandler,   /* data ref. handler */
        ConstStr255Param fileName)                    /* file name for extension */
{
    OSStatus anErr = noErr;
    unsigned char myChar = 0;
    Handle fileNameHndl = NULL;

    /* create a handle with our file name string */

    /* if we were passed a null string, then we
       need to add this null string (a single 0
       byte) to the handle */

    if (fileName == NULL)
        anErr = PtrToHand(&myChar, &fileNameHndl, sizeof(myChar));
    else
        anErr = PtrToHand(fileName, &fileNameHndl, fileName[0] + 1);
    if (anErr != noErr) goto bail;

    /* set the data ref extension for the
        data ref handler */
    anErr = DataHSetDataRefExtension(
            dataRefHandler,         /* data ref. handler */
            fileNameHndl,           /* data ref. extension to add */
            kDataRefExtensionFileName);

bail:
    if (fileNameHndl)
        /* no longer need this */
        DisposeHandle(fileNameHndl);

    return anErr;

}

//////////
//
// PtrDataRef_AddFileTypeExtension
//
// Tell the data handler to set
// the file type extension in the
// data reference.
//
//////////

static OSStatus PtrDataRef_AddFileTypeExtension(
        ComponentInstance dataRefHandler,   /* data ref. handler */
        OSType fileType)                    /* file type for extension */
{
    Handle      fileTypeHndl = NULL;
    OSStatus    anErr        = noErr;
    OSType      myType;

    myType = EndianU32_NtoB(fileType);

    anErr = PtrToHand(&myType, &fileTypeHndl, sizeof(OSType));
    if (anErr != noErr) goto bail;

    /* set the data ref extension for the
        data ref handler */
    anErr = DataHSetDataRefExtension(
            dataRefHandler,             /* data ref. handler */
            fileTypeHndl,               /* data ref. extension to add */
            kDataRefExtensionMacOSFileType);

bail:

    if (fileTypeHndl)
        /* no longer need this */
        DisposeHandle(fileTypeHndl);

    return anErr;
}

//////////
//
// PtrDataRef_AddMIMETypeExtension
//
// Tell the data handler to set
// the mime type extension in the
// data reference.
//
//////////

static OSStatus PtrDataRef_AddMIMETypeExtension(
        ComponentInstance dataRefHandler,   /* data ref. handler */
        StringPtr mimeType)                 /* mime type for extension */
{
    OSStatus anErr = noErr;
    Handle mimeTypeHndl = NULL;

    if (mimeType == NULL)
        return paramErr;

    anErr = PtrToHand(mimeType, &mimeTypeHndl, mimeType[0] + 1);
    if (anErr != noErr) goto bail;

    /* set the data ref extension for the
        data ref handler */
    anErr = DataHSetDataRefExtension(
            dataRefHandler,             /* data ref. handler */
            mimeTypeHndl,               /* data ref. extension to add */
            kDataRefExtensionMIMEType);

bail:

    if (mimeTypeHndl)
        /* no longer need this */
        DisposeHandle(mimeTypeHndl);

    return anErr;
}


//////////
//
// MyCreatePointerReferenceHandle
// Create a pointer data reference handle.
//
// The pointer data reference handle contains
// a record specifying a pointer to a block of
// movie data along with a size value.
//
//////////

static Handle MyCreatePointerReferenceHandle(void *data, Size dataSize)
{
	Handle dataRef = NULL;
	PointerDataRefRecord ptrDataRefRec;
	OSErr err;


	ptrDataRefRec.data = data;
	ptrDataRefRec.dataLength = dataSize;

	// create a data reference handle for our data
	err = PtrToHand( &ptrDataRefRec, &dataRef, sizeof(PointerDataRefRecord));

	return dataRef;
}

Handle QT_createPointerDataRefWithExtensions(
                         void               *data,
                         Size               dataSize,
                         ConstStr255Param   fileName,
                         OSType             fileType,
                         StringPtr          mimeTypeString
                         )
{
    OSStatus  err = noErr;
    Handle dataRef = NULL;
    ComponentInstance dataRefHandler = NULL;

    // First create a data reference handle for our data
    dataRef = MyCreatePointerReferenceHandle(data, dataSize);
    if (!dataRef) goto bail;

    //  Get a data handler for our data reference
    err = OpenADataHandler(
            dataRef,                    /* data reference */
            PointerDataHandlerSubType,  /* data ref. type */
            NULL,                       /* anchor data ref. */
            (OSType)0,                  /* anchor data ref. type */
            NULL,                       /* time base for data handler */
            kDataHCanRead,              /* flag for data handler usage */
            &dataRefHandler);           /* returns the data handler */
    if (err) goto bail;

    // We can add the filename to the data ref to help
    // importer finding process. Find uses the extension.
    // If we add a filetype or mimetype we must add a
    // filename -- even if it is an empty string
    if (fileName || fileType || mimeTypeString)
    {
        err = PtrDataRef_AddFileNameExtension(
            dataRefHandler, /* data ref. handler */
            fileName);      /* file name for extension */

        if (err) goto bail;
    }

    // The pointer data handler can also be told the
    // filetype and/or MIME type by adding data ref
    // extensions. These help the importer finding process.
    // NOTE: If you add either of these, you MUST add
    // a filename first -- even if it is an empty Pascal
    // string. Any data ref extensions will be ignored.

    // to add file type, you add a classic atom followed
    // by the Mac OS filetype for the kind of file

    if (fileType)
    {
        err = PtrDataRef_AddFileTypeExtension(
            dataRefHandler, /* data ref. handler */
            fileType);      /* file type for extension */
        if (err) goto bail;
    }


    // to add MIME type information, add a classic atom followed by
    // a Pascal string holding the MIME type

    if (mimeTypeString)
    {
        err = PtrDataRef_AddMIMETypeExtension (
                dataRefHandler,     /* data ref. handler */
                mimeTypeString);    /* mime string for extension */
        if (err) goto bail;
    }

    /* dispose old data ref handle because
        it does not contain our new changes */
    DisposeHandle(dataRef);
    dataRef = NULL;

    /* re-acquire data reference from the
        data handler to get the new
        changes */
    err = DataHGetDataRef(dataRefHandler, &dataRef);
    if (err) goto bail;

    CloseComponent(dataRefHandler);

    return dataRef;

bail:
    if (dataRefHandler)
    {
        CloseComponent(dataRefHandler);
    }

    if (dataRef)
    {
        // make sure and dispose the data reference handle
        // once we are done with it
        DisposeHandle(dataRef);
    }

    return NULL;
}

void PlayMemory_QT(char *inMem, const int inSize)
{
	movieSound_t *newSound = NULL;

	newSound = (movieSound_t *)calloc (sizeof(movieSound_t), 1);
	newSound->mLooping = false;
	newSound->mAutoDelete = true;
	newSound->mIsMusic = false;

	// Open the movie
	newSound->mDataRef = QT_createPointerDataRefWithExtensions(inMem, inSize, "\ptest.wav", NULL, NULL);
	require_noerr( NewMovieFromDataRef(&(newSound->mMovie), newMovieDontAskUnresolvedDataRefs, NULL, newSound->mDataRef, PointerDataHandlerSubType), cantCreateMovie);

	// This is an audio-only movie, so disassociate any visual contexts from it
	SetMovieVisualContext (newSound->mMovie, NULL);
	
	// Go to the beginning of the movie
	SetMovieActive(newSound->mMovie, true);
	GoToBeginningOfMovie(newSound->mMovie);
	PrerollMovie(newSound->mMovie, 0, 0);

	SetupLoopCallback(newSound);
	
	// Start playback
	StartMovie(newSound->mMovie);
	newSound->mPlaying = true;

cantCreatePath:
cantCreateDataRef:
cantCreateMovie:
	if (newSound && newSound->mDataRef)
	{
		DisposeHandle(newSound->mDataRef);
		newSound->mDataRef = NULL;
	}
}

//*********************************************************
void StopMusic()
{
	StopSound_QT(sMusicSound);
	sMusicSound = NULL;
	MusicPlaying = 0;
}

//****************************************************************************
void PlayMusic(char *pMem)
{
#if _DEBUG && DEBUG_SOUND
printf("PlayMusic: %s\n", pMem);
#endif
	BuildPath(pMem, DIR_MUSIC, FALSE);
	StopMusic();
	sMusicSound = PlayFile_QT (TmpPath, false, false, true);
	if (sMusicSound) MusicPlaying = (WORD)sMusicSound->mPlaying;
}

//*******************************************************************
void KPlaySound(char *pMem)
{
#if _DEBUG && DEBUG_SOUND
printf("KPlaySound: %s\n", pMem);
#endif
	BuildPath(pMem, DIR_SOUNDS, FALSE);
	(void)PlayFile_QT (TmpPath, false, true, false);
}

//*******************************************************************
void KPlaySoundMem(char *pMem, BOOL nostop)
{
	// LBO 12/13/07. Used in the MESH games, e.g. when a rock drops
	// TODO: act on nostop? If false, it's supposed to not play a sound if one is already playing
	int waveSize = *((DWORD *)pMem+1);
	waveSize = EndianS32_LtoN(waveSize) + 8;
	(void)PlayMemory_QT(pMem, waveSize);
}

//******************************************************************
void KGetOSversion(BYTE *pMem) {

	SInt32 majorVersion, minorVersion, bugFixVersion;

	if ((Gestalt(gestaltSystemVersionMajor, &majorVersion) == noErr) &&
		(Gestalt(gestaltSystemVersionMinor, &minorVersion) == noErr) &&
		(Gestalt(gestaltSystemVersionBugFix, &bugFixVersion) == noErr))
	{
		sprintf((char *)pMem, "Mac OS X %d.%d.%d", majorVersion, minorVersion, bugFixVersion);
		// as written, this should handle up to OS X 99.99.99
//		gOSXVersion32 = ((majorVersion / 10) << 20) |
//		((majorVersion % 10) << 16) |
//		((minorVersion / 10) << 12) |
//		((minorVersion % 10) << 8) |
//		((bugFixVersion / 10) << 4) |
//		((bugFixVersion % 10) << 0);
	} else strcpy((char *)pMem, "Mac OS X");
	
	// We require 10.4.0 and later.
	if (gOSXVersion32 < 0x100400)
	{
		SystemMessage("This game requires Mac OS X 10.4.0 or later.");
		exit(0);
	}
	
}

//***************
//* END OF FILE *
//***************
